Type: Package
Package: duckh3
Title: H3 Extension of 'DuckDB'
Version: 0.1.0
Authors@R: c(
    person(
        "Adrián", "Cidre González", 
        email = "adrian.cidre@gmail.com", 
        role = c("aut", "cre"),
        comment = c(ORCID = "0000-0002-3310-3052")
    ),
    person(
        "Egor", "Kotov", , 
        "kotov.egor@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0001-6690-5345")
    ),
    person(
        "Rafael H. M.", "Pereira", , 
        "rafa.pereira.br@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0003-2125-7465")
    ))
Description: Fast & memory-efficient functions to analyze and manipulate large 
            data sets. It leverages the fast analytical 
            capabilities of 'DuckDB' and its spatial extension (see <https://duckdb.org/community_extensions/extensions/h3>) 
            while maintaining compatibility with R’s spatial data ecosystem to 
            work with spatial vector data.
URL: https://cidree.github.io/duckh3/, https://github.com/Cidree/duckh3
BugReports: https://github.com/Cidree/duckh3/issues
Encoding: UTF-8
License: GPL (>= 3)
Imports: cli, DBI, dbplyr, dplyr, duckdb, duckspatial (>= 1.0.0), glue
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Suggests: sf, testthat (>= 3.0.0)
NeedsCompilation: no
Packaged: 2026-04-22 18:10:08 UTC; Cidre
Author: Adrián Cidre González [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3310-3052>),
  Egor Kotov [aut] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Rafael H. M. Pereira [aut] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>)
Maintainer: Adrián Cidre González <adrian.cidre@gmail.com>
Depends: R (>= 4.1.0)
Repository: CRAN
Date/Publication: 2026-04-24 20:10:10 UTC
Built: R 4.5.2; ; 2026-04-24 20:47:31 UTC; unix

Type: Package
Package: duckh3
Title: H3 Extension of 'DuckDB'
Version: 0.1.0
Authors@R: c(
    person(
        "Adrián", "Cidre González", 
        email = "adrian.cidre@gmail.com", 
        role = c("aut", "cre"),
        comment = c(ORCID = "0000-0002-3310-3052")
    ),
    person(
        "Egor", "Kotov", , 
        "kotov.egor@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0001-6690-5345")
    ),
    person(
        "Rafael H. M.", "Pereira", , 
        "rafa.pereira.br@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0003-2125-7465")
    ))
Description: Fast & memory-efficient functions to analyze and manipulate large 
            data sets. It leverages the fast analytical 
            capabilities of 'DuckDB' and its spatial extension (see <https://duckdb.org/community_extensions/extensions/h3>) 
            while maintaining compatibility with R’s spatial data ecosystem to 
            work with spatial vector data.
URL: https://cidree.github.io/duckh3/, https://github.com/Cidree/duckh3
BugReports: https://github.com/Cidree/duckh3/issues
Encoding: UTF-8
License: GPL (>= 3)
Imports: cli, DBI, dbplyr, dplyr, duckdb, duckspatial (>= 1.0.0), glue
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Suggests: sf, testthat (>= 3.0.0)
NeedsCompilation: no
Packaged: 2026-04-22 18:10:08 UTC; Cidre
Author: Adrián Cidre González [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3310-3052>),
  Egor Kotov [aut] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Rafael H. M. Pereira [aut] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>)
Maintainer: Adrián Cidre González <adrian.cidre@gmail.com>
Depends: R (>= 4.1.0)
Repository: CRAN
Date/Publication: 2026-04-24 20:10:10 UTC
Built: R 4.6.0; ; 2026-04-29 15:53:27 UTC; unix

# Generated by roxygen2: do not edit by hand

export(ddbh3_bigint_to_strings)
export(ddbh3_create_conn)
export(ddbh3_default_conn)
export(ddbh3_get_center_child)
export(ddbh3_get_child_pos)
export(ddbh3_get_children)
export(ddbh3_get_icosahedron_faces)
export(ddbh3_get_n_children)
export(ddbh3_get_parent)
export(ddbh3_get_resolution)
export(ddbh3_h3_to_lat)
export(ddbh3_h3_to_lon)
export(ddbh3_h3_to_points)
export(ddbh3_h3_to_spatial)
export(ddbh3_h3_to_vertex)
export(ddbh3_h3_to_vertexes)
export(ddbh3_is_h3)
export(ddbh3_is_pentagon)
export(ddbh3_is_res_class_iii)
export(ddbh3_is_vertex)
export(ddbh3_lonlat_to_h3)
export(ddbh3_lonlat_to_spatial)
export(ddbh3_points_to_h3)
export(ddbh3_points_to_spatial)
export(ddbh3_strings_to_bigint)
export(ddbh3_vertex_to_lat)
export(ddbh3_vertex_to_lon)
export(ddbh3_vertex_to_spatial)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# duckh3 <a href="https://cidree.github.io/duckh3/"><img src="man/figures/logo.png" alt="duckh3 website" align="right" height="138"/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/duckh3)](https://CRAN.R-project.org/package=duckh3)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![Codecov test
coverage](https://codecov.io/gh/Cidree/duckh3/graph/badge.svg)](https://app.codecov.io/gh/Cidree/duckh3)
[![License: GPL
v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![check](https://github.com/Cidree/duckh3/workflows/check/badge.svg)](https://github.com/Cidree/duckh3/actions)

<!-- badges: end -->

**{duckh3}** provides fast, memory-efficient functions for analysing and
manipulating large spatial and non-spatial datasets using the [H3
hierarchical indexing system](https://h3geo.org/) in R. It bridges
[DuckDB’s H3
extension](https://duckdb.org/community_extensions/extensions/h3) with
R’s data and spatial ecosystems — in particular **{duckspatial}**,
**{dplyr}**, and **{sf}** — so you can leverage DuckDB’s analytical
power without leaving your familiar R workflow.

### How it works

{duckh3} was built around the same API as the [duckspatial R
package](https://github.com/Cidree/duckspatial), and it operates on
regular R data frames, `tibble`s, `dbplyr` lazy tables, and
`duckspatial_df` objects. Unlike purely spatial workflows, H3 operations
do not require your data to be spatial. Any table with
longitude/latitude columns, or an existing H3 index column, is a valid
starting point.

When a DuckDB connection is used, all H3 operations run inside that
connection with the H3 extension enabled, letting DuckDB apply its own
query optimisations before any data reaches R. Results are returned
lazily and only materialised when you explicitly collect them.

In addition, {duckh3} registers a set of **DuckDB macros** on the
default connection at load time, making H3 functions available directly
inside `dplyr::mutate()` on lazy tables — no wrapper function needed.
Note that they **only work with lazy tables**, not with regular data
frames.

### Naming conventions

All functions follow the `ddbh3_*()` prefix (*DuckDB H3*), structured
around the expected input data, and what they will be converted to:

- `ddbh3_lonlat_to_*()` — from longitude/latitude coordinates to H3
  representations
- `ddbh3_points_to_*()` — from spatial point geometries to H3
  representations
- `ddbh3_h3_to_*()` — convert H3 cells to other representations
- `ddbh3_vertex_to_*()` — convert H3 vertexes to other representations

With the following available transformations:

| Function family  | Output                             |
|------------------|------------------------------------|
| `*_to_h3()`      | H3 index as string or `UBIGINT`    |
| `*_to_spatial()` | H3 cell as spatial hexagon polygon |
| `*_to_lon()`     | Longitude of H3 cell centroid      |
| `*_to_lat()`     | Latitude of H3 cell centroid       |

And there are also a set of function to retrieve or check properties of
the data:

- `ddbh3_get_*()` — retrieve H3 cell properties (resolution, parent,
  children, vertices…)
- `ddbh3_is_*()` — check properties of H3 indexes (valid, pentagon,
  Class III…)

## Installation

Install the stable release from CRAN:

``` r
pak::pak("duckh3")
```

Install the latest GitHub version (more features, fewer accumulated
bugs):

``` r
# install.packages("pak")
pak::pak("Cidree/duckh3")
```

Install the development version (may be unstable):

``` r
pak::pak("Cidree/duckh3@dev")
```

## Contributing

Bug reports, feature requests, and pull requests are very welcome!

- [Raise an issue](https://github.com/Cidree/duckh3/issues)
- [Open a pull request](https://github.com/Cidree/duckh3/pulls)

#' Get the resolution of H3 cell indexes
#'
#' Get the resolution of H3 cell indexes stored as strings or unsigned 64-bit
#' integers (`UBIGINT`) from an existing column.
#'
#' @template x
#' @template h3
#' @template new_column
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#'
#' @template returns_tbl
#' @export
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(duckspatial)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add h3 strings
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 10)
#' 
#' ## Convert to duckspatial_df
#' points_ddbs <- ddbs_as_points(points_tbl)
#' 
#' ## Get resolution of the h3 strings
#' ddbh3_get_resolution(points_tbl)
#' ddbh3_get_resolution(points_ddbs, new_column = "res")
#' 
#' ## Add using mutate
#' points_tbl |> 
#'   mutate(res = ddbh3_get_resolution(h3string))
ddbh3_get_resolution <- function(
    x,
    h3 = "h3string",
    new_column = "h3resolution",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_character_scalar(new_column, "new_column")


  # 1. Build parameters string
  built_fun <- glue::glue("h3_get_resolution({h3})")


  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_get_resolution(x)"
  ) 

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## create duckdb connection
conn_test <- duckh3::ddbh3_create_conn()

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)


# 1. ddbs_get_resolution() -----------------------------------------------

## 1.1. Input data in different formats ----------

testthat::describe("ddbh3_get_resolution() works in different formats", {

  ## Get h3strings
  test_data <- ddbh3_lonlat_to_h3(points_tbl, resolution = 10)

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct resolution for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_resolution(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check resolution is the same as created
    res_col <- dplyr::collect(res)
    expect_all_equal(res_col$h3resolution, 10)
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct resolution for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_get_resolution(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check resolution is the same as created
    res_col <- dplyr::collect(res)
    expect_all_equal(res_col$h3resolution, 10)
  })


  ## FORMAT 3 - SF
  testthat::it("returns the correct resolution for sf", {
    ## Convert to sf
    test_data_sf <- test_data |> 
      dplyr::collect() |> 
      sf::st_as_sf(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_resolution(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check resolution is the same as created
    res_col <- duckspatial::ddbs_collect(res)
    expect_all_equal(res_col$h3resolution, 10)
  })


  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct resolution for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |> 
      dplyr::collect() |> 
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_resolution(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check resolution is the same as created
    res_col <- duckspatial::ddbs_collect(res)
    expect_all_equal(res_col$h3resolution, 10)
  })


  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct resolution for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to duckspatial_df
    test_data_sf <- test_data |> 
      dplyr::collect() |> 
      sf::st_as_sf(coords = c("lon", "lat"), crs = 4326)
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_resolution("sf_pts", conn = conn_test)
    ## Check
    expect_s3_class(res, "duckspatial_df")
    ## Check resolution is the same as created
    res_col <-duckspatial::ddbs_collect(res)
    expect_all_equal(res_col$h3resolution, 10)
  })
  
  
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_get_resolution() arguments work", {
  
  ## Get h3strings
  test_data <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)
  
  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |> 
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_resolution(test_data_mod, h3 = "h3")
    ## Checks
    expect_true("h3resolution" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_all_equal(res_col$h3resolution, 8)
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_resolution(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_all_equal(res_col$res, 8)
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_resolution(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl"
    ))
    ## Apply operation with connection
    expect_no_message(ddbh3_get_resolution(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_resolution(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_resolution(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl",
      overwrite = TRUE
    ))

  })
  
})

## 1.3. Errors on weird inputs -----------


describe("errors", {

  ## Get h3strings
  test_data <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))
    
  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_resolution(test_data, h3 = NULL))
    expect_error(ddbh3_get_resolution(test_data, h3 = TRUE))
    expect_error(ddbh3_get_resolution(test_data, h3 = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_resolution(test_data, new_column = NULL))
    expect_error(ddbh3_get_resolution(test_data, new_column = FALSE))
    expect_error(ddbh3_get_resolution(test_data, new_column = 25))
  })
  
  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_get_resolution("test_data_tbl", conn = NULL)))
    )
  })
  
  it("validates x argument type", {
    expect_error(ddbh3_get_resolution(x = 999))
  })
  
  it("validates conn argument type", {
    expect_error(ddbh3_get_resolution(test_data, conn = 999))
  })
  
  it("validates overwrite argument type", {
    expect_error(ddbh3_get_resolution(test_data, overwrite = 999))
  })
  
  it("validates quiet argument type", {
    expect_error(ddbh3_get_resolution(test_data, quiet = 999))
  })
  
  it("validates table name exists", {
    expect_error(ddbh3_get_resolution(x = "999", conn = conn_test))
  })
  
  it("requires name to be single character string", {
    expect_error(ddbh3_get_resolution(test_data, conn = conn_test, name = c('banana', 'banana')))
  })
})
  

#' Template for main h3 operations
#'
#' @template x
#' @template new_column
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#' @param fun The duckdb function to use
#' @param other_params string with other function-specific parameters
#' 
#' @keywords internal
#' @noRd
template_h3_base <- function(
    x,
    new_column = "h3string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE,
    fun = NULL,
    base_fun = NULL
) {

  ## First, try vectorized option
  if (is.null(conn) && (is.character(x) | is.numeric(x)) && !is.null(base_fun)) {
    res <- get_vectorized_result(x, base_fun)
    return(res)
  }

  # 0. Validate inputs
  duckspatial:::assert_character_scalar(new_column, "new_column")
  duckspatial:::assert_conn_character(conn, x)
  duckspatial:::assert_name(name)
  duckspatial:::assert_logic(overwrite, "overwrite")
  duckspatial:::assert_logic(quiet, "quiet")


  # 1. Prepare inputs

  ## 1.1. Pre-extract CRS  
  sf_col_x <- attr(x, "sf_column")
  
  ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- suppressWarnings(duckspatial:::normalize_spatial_input(x, conn))

  ## 1.3. Extract CRS
  crs_x    <- tryCatch(
    suppressWarnings(duckspatial::ddbs_crs(x, conn)), 
    error = function(e) NULL
  )

  ## 1.4. Resolve spatial connections and handle imports
  resolve_conn <- duckspatial:::resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.5. Get list with query names for the input data
  x_list <- duckspatial:::get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  ## 1.6. Install and load h3 on target connection
  ## TODO - add argument in duckspatial:::resolve_spatial_connections() to manage it
  # duckspatial::ddbs_install(target_conn, upgrade = FALSE, quiet = TRUE, extension = "h3")
  # duckspatial::ddbs_load(target_conn, quiet = TRUE, extension = "h3")


  # 2. Prepare the query

  ## 2.1. Get names of geometry columns (use saved sf_col_x from before transformation)
  ## since the input of this function can be a non-spatial data frame, x_geom can return
  ## an empty value
  x_geom <- sf_col_x %||% duckspatial:::get_geom_name(target_conn, x_list$query_name)
  if (length(x_geom != 0)) duckspatial:::assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the base query
  geom_clause <- if (length(x_geom) > 0) {
    glue::glue("
      * EXCLUDE {x_geom},
      {fun} as {new_column},
      {duckspatial:::build_geom_query(x_geom, name, crs_x, 'duckspatial')} AS {x_geom}"
    )
  } else {
    glue::glue("*, {fun} as {new_column}")
  }
  base.query <- glue::glue("
    SELECT
      {geom_clause}
    FROM
      {x_list$query_name};
  ")

  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    duckspatial:::create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
   duckspatial::: ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = "duckspatial",
      crs    = crs_x,
      x_geom = x_geom
    )
  }

}



template_h3_to_spatial <- function(
    x,
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE,
    fun = NULL,
    base_fun = NULL
) {

  # TODO
  # First, try vectorized option
  if (is.null(conn) && is.character(x) && !is.null(base_fun)) {
    res <- get_vectorized_result(x, base_fun)
    return(res)
  }

  # 0. Validate inputs
  duckspatial:::assert_conn_character(conn, x)
  duckspatial:::assert_name(name)
  duckspatial:::assert_logic(overwrite, "overwrite")
  duckspatial:::assert_logic(quiet, "quiet")

  ## 1.1. Pre-extract CRS  
  sf_col_x <- attr(x, "sf_column")
  
  ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- suppressWarnings(duckspatial:::normalize_spatial_input(x, conn))

  ## 1.3. Extract CRS
  crs_x    <- tryCatch(
    suppressWarnings(duckspatial::ddbs_crs(x, conn)), 
    error = function(e) NULL
  )

  ## 1.4. Resolve spatial connections and handle imports
  resolve_conn <- duckspatial:::resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.5. Get list with query names for the input data
  x_list <- duckspatial:::get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  ## 1.6. Install and load h3 on target connection
  # duckspatial::ddbs_install(target_conn, upgrade = FALSE, quiet = TRUE, extension = "h3")
  # duckspatial::ddbs_load(target_conn, quiet = TRUE, extension = "h3")


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% duckspatial:::get_geom_name(target_conn, x_list$query_name)
  if (length(x_geom != 0)) duckspatial:::assert_geometry_column(x_geom, x_list)

  ## 3.3. Build the columns clause
  if (length(x_geom) == 0) {
    x_geom <- "geometry"
    cols_clause <- glue::glue("*, {duckspatial:::build_geom_query(fun, name, 'EPSG:4326', 'duckspatial')} AS geometry")
  } else {
    cols_clause <- glue::glue("* REPLACE ({duckspatial:::build_geom_query(fun, name, crs_x, 'duckspatial')} AS {x_geom})")
  }

  ## 3.3. Build the base query
  base.query <- glue::glue("
    SELECT
      {cols_clause}
    FROM
      {x_list$query_name};
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    duckspatial:::create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    duckspatial:::ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = "duckspatial",
      crs    = "EPSG:4326",
      x_geom = x_geom
    )
  }

}

#' @param memory_limit_gb Numeric. Memory limit in GB. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use 80% of available operating system memory it detects (warning, on some shared HPC nodes the detected memory might be the full node memory, not the per-job allocation).

#' Convert H3 string or UBIGINT indexes to other representations
#'
#' Convert H3 cell indexes stored as strings or UBIGINT into other 
#' representations (e.g. `lon`, `lat`, `spatial`)
#'
#' @template x
#' @template h3
#' @template new_column
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#'
#' @details
#' The four functions differ only in the output format:
#' * `ddbh3_h3_to_spatial()` converts H3 indexes to spatial hexagon polygons
#' * `ddbh3_h3_to_lon()` extracts the longitude of the H3 cell centroid
#' * `ddbh3_h3_to_lat()` extracts the latitude of the H3 cell centroid
#' * `ddbh3_strings_to_bigint()` converts H3 indexes to unsigned 64-bit integers (`UBIGINT`)
#' * `ddbh3_bigint_to_strings()` converts H3 indexes to strings (e.g. `"8928308280fffff"`)
#' * `ddbh3_h3_to_points()` converts H3 indexes to spatial points located at the centroid 
#' of the H3 cell
#'
#' @template desc_formats
#'
#' @template returns_tbl
#'
#' @name ddbh3_h3_to
#' @rdname ddbh3_h3_to
#' @aliases ddbh3_h3_to_spatial ddbh3_h3_to_lon ddbh3_h3_to_lat ddbh3_strings_to_bigint ddbh3_bigint_to_strings
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(duckspatial)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add H3 string column
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 6) |> 
#'   select(-lon, -lat)
#' 
#' ## TO LON/LAT ------------
#' 
#' ## Add longitude and latitude of the H3 string
#' points_coords_tbl <- points_tbl |> 
#'   ddbh3_h3_to_lat() |> 
#'   ddbh3_h3_to_lon()
#' 
#' ## Add lon/lat with other names
#' points_coords_2_tbl <- points_tbl |> 
#'   ddbh3_h3_to_lat(new_column = "latitude") |> 
#'   ddbh3_h3_to_lon(new_column = "longitude")
#' 
#' ## Add using mutate
#' points_tbl |> 
#'   mutate(
#'     lon = ddbh3_h3_to_lon(h3string),
#'     lat = ddbh3_h3_to_lat(h3string)
#'   )
#' 
#' ## TO SPATIAL -----------------
#' 
#' ## Convert h3 strings to spatial polygons
#' points_ddbs <- ddbh3_h3_to_spatial(points_tbl)
#' 
#' ## Collect as sf
#' points_sf <- ddbs_collect(points_ddbs)
#' 
#' ## FROM STRING TO UBIGINT -----
#' 
#' ## Add ubigint, and remove strings
#' points_bigint_tbl <- ddbh3_strings_to_bigint(
#'   points_tbl, 
#'   new_column = "h3_integers"
#' ) |> 
#'   select(-h3string)
#' 
#' ## Add using mutate
#' points_tbl |> 
#'   mutate(h3int = ddbh3_strings_to_bigint(h3string))
#' 
#' ## FROM UBIGINT TO STRING -----
#' 
#' ## Add column with strings
#' points_strings_tbl <- ddbh3_bigint_to_strings(
#'   points_bigint_tbl, 
#'   h3 = "h3_integers"
#' ) 
#' 
#' ## Add using mutate
#' points_bigint_tbl |> 
#'   mutate(h3string = ddbh3_bigint_to_strings(h3_integers))
#' 
NULL





#' @rdname ddbh3_h3_to
#' @export
ddbh3_h3_to_lon <- function(
    x,
    h3 = "h3string",
    new_column = "lon",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_lng({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_cell_to_lng(x)"
  ) 

}





#' @rdname ddbh3_h3_to
#' @export
ddbh3_h3_to_lat <- function(
    x,
    h3 = "h3string",
    new_column = "lat",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_lat({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_cell_to_lat(x)"
  ) 

}





#' @rdname ddbh3_h3_to
#' @export
ddbh3_h3_to_spatial <- function(
    x,
    h3 = "h3string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {

  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("
    ST_GeomFromWKB(
      h3_cell_to_boundary_wkb({h3})
    )"
  )

  # 2. Pass to template
  template_h3_to_spatial(
    x = x,
    conn = conn,
    name = name,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "ST_GeomFromWKB(h3_cell_to_boundary_wkb(x))"
  ) 

}





#' @rdname ddbh3_h3_to
#' @export
ddbh3_strings_to_bigint <- function(
    x,
    h3 = "h3string",
    new_column = "h3bigint",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {


  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_string_to_h3({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_string_to_h3(x)"
  ) 

}



#' @rdname ddbh3_h3_to
#' @export
ddbh3_bigint_to_strings <- function(
    x,
    h3 = "h3bigint",
    new_column = "h3string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {


  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_h3_to_string({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_h3_to_string(x)"
  ) 

}




#' @rdname ddbh3_h3_to
#' @export
ddbh3_h3_to_points <- function(
    x,
    h3 = "h3string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("
    ST_Point(
      h3_cell_to_lng({h3}),
      h3_cell_to_lat({h3})
    )
  ")

  # 2. Pass to template
  template_h3_to_spatial(
    x = x,
    conn = conn,
    name = name,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "ST_Point(
      h3_cell_to_lng(x), 
      h3_cell_to_lat(x)
    )"
  ) 

}

#' Create a DuckDB connection with spatial and h3 extensions
#'
#' It creates a DuckDB connection, and then it installs and loads the
#' spatial and h3 extensions
#'
#' @param dbdir String. Either "tempdir", "memory", or file path with .duckdb 
#' or .db extension. Defaults to "memory".
#' @template threads
#' @template memory_limit_gb
#' @param bigint String. How to handle 64-bit integers. One of "integer64" or "numeric"
#' @param ... Other parameters passed to `DBI::dbConnect()`
#'
#' @returns A `duckdb_connection`
#' @export
#'
#' @examples
#' \donttest{
#' # load packages
#' library(duckspatial)
#' library(duckh3)
#'
#' # create a duckdb database in memory
#' conn <- ddbh3_create_conn(dbdir = "memory", threads = 1)
#' 
#' # create an in-memory connection with 1 thread and 2GB memory limit
#' conn <- ddbh3_create_conn(threads = 1, memory_limit_gb = 2)
#'
#' # Create a persistent database in disk
#' # conn <- ddbh3_create_conn(dbdir = "my_database.duckdb")
#'
#' ddbs_stop_conn(conn)
#' }
ddbh3_create_conn <- function(
  dbdir = "memory", 
  threads = NULL, 
  memory_limit_gb = NULL,
  bigint = "integer64",
  ...
) {

  # Creates a connection with the Spatial extension
  conn <- duckspatial::ddbs_create_conn(
    dbdir = dbdir,
    threads = threads,
    memory_limit_gb = memory_limit_gb,
    bigint = bigint,
    ...
  )
  
  # Checks and installs the h3 extension
  duckspatial::ddbs_install(conn, upgrade = FALSE, quiet = TRUE, extension = "h3")
  duckspatial::ddbs_load(conn, quiet = TRUE, extension = "h3")

  return(conn)
}




#' Get or create default DuckDB connection
#' 
#' Setup the default connection with h3 and spatial extensions installed and loaded.
#' It will be used internally by the package functions if no other connection is provided
#'
#' @param create Logical. If TRUE and no connection exists, create one.
#' Default is TRUE.
#' @param upgrade_h3 Logical. If TRUE, will attempt to upgrade the h3 extension
#' @param ... Additional parameters to pass to [duckspatial::ddbs_create_conn()]
#'
#' @returns Invisibly, the default `duckdb_connection`
#'
#' @export
#' @examples
#' # Get or create default connection
ddbh3_default_conn <- function(
  create = TRUE,
  upgrade_h3 = FALSE,
  ...
) {
  conn <- getOption("duckspatial_conn", NULL)

  # Check if existing connection is still valid

  if (!is.null(conn)) {
    if (!DBI::dbIsValid(conn)) {
      options(duckspatial_conn = NULL)
      conn <- NULL
    }
  }

  # Create new connection if needed
  if (is.null(conn) && create) {
    conn <- duckspatial::ddbs_create_conn(
      dbdir = "memory", 
      bigint = "integer64",
      ...
    )
    options(duckspatial_conn = conn)
  }

  # Install and load h3 extension
  duckspatial::ddbs_install(conn, upgrade = upgrade_h3, quiet = TRUE, extension = "h3")
  duckspatial::ddbs_load(conn, quiet = TRUE, extension = "h3")

  # Create macros in the default connection
  create_ddbh3_default_macros()

  invisible(conn)
}

#' String representations are the most portable and human-readable. `UBIGINT`
#' indexes are more memory-efficient and better suited for large-scale joins or
#' arithmetic operations. Spatial representations are useful for direct
#' visualization or spatial operations within R's spatial ecosystem.

#' Convert spatial points to H3 cell representations
#'
#' Convert spatial point geometries into H3 cell representations at a specified
#' resolution. Input must be an `duckspatial_df`, `sf`, or table with point
#' geometries
#'
#' @template x
#' @template resolution
#' @template new_column
#' @param h3_format Character. Output format for the H3 cell index. Either
#' `"string"` (default) or `"bigint"`. Only used in `ddbh3_points_to_h3()`.
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#' 
#' @details
#' The two functions differ in the output format:
#' * `ddbh3_points_to_h3()` returns the H3 cell index containing each point,
#'   either as a string or `UBIGINT` depending on `h3_format`
#' * `ddbh3_points_to_spatial()` returns the H3 cell hexagon polygon containing
#'   each point as a spatial geometry
#'
#' @template returns_tbl
#'
#' @name ddbh3_points_to
#' @aliases ddbh3_points_to_h3 ddbh3_points_to_spatial
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(duckspatial)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Convert to duckspatial_df
#' points_ddbs <- ddbs_as_points(points_tbl)
#' 
#' ## TO H3 strings/ubigint ------------
#' 
#' ## Add column with h3 strings at resolution 8
#' points_strings_ddbs <- ddbh3_points_to_h3(points_ddbs, 8)
#' 
#' ## Add column with h3 ubigint at resolution 10
#' points_bigint_ddbs <- ddbh3_points_to_h3(
#'   points_ddbs, 
#'   resolution = 8,
#'   new_column = "h3bigint",
#'   h3_format = "bigint"
#' )
#' 
#' ## TO SPATIAL -----------------
#' 
#' ## Convert from POINTS to H3 POLYGONS of res 8
#' polygons_8_ddbs <- ddbh3_points_to_spatial(points_ddbs, 8)
#' 
#' ## Collect as sf
#' polygons_8_sf <- ddbs_collect(polygons_8_ddbs)
NULL





#' @rdname ddbh3_points_to
#' @export
ddbh3_points_to_spatial <- function(
    x,
    resolution,
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {

  # 0. Validate inputs
  duckspatial:::assert_conn_character(conn, x)
  duckspatial:::assert_name(name)
  duckspatial:::assert_logic(overwrite, "overwrite")
  duckspatial:::assert_logic(quiet, "quiet")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")
  duckspatial:::assert_geom_type(x = x, conn = conn, geom = "POINT", multi = TRUE)

  # 1. Manage connection to DB

  ## 1.1. Pre-extract geometry column name to overwrite, if present
  crs_x    <- duckspatial::ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")

  if (crs_x$input != "EPSG:4326") {
    cli::cli_abort("The CRS of the input must be {.val EPSG:4326}, not {.val crs_x$input}.")
  }


  ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df,
  ## validate character table names
  x <- suppressWarnings(duckspatial:::normalize_spatial_input(x, conn))

  # 2. Manage connection to DB

  ## 2.1. Resolve connections and handle imports
  resolve_conn <- duckspatial:::resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 2.2. Get query list of table names
  x_list <- duckspatial:::get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  ## 2.3. Install and load h3 on target connection
  # duckspatial::ddbs_install(target_conn, upgrade = FALSE, quiet = TRUE, extension = "h3")
  # duckspatial::ddbs_load(target_conn, quiet = TRUE, extension = "h3")


  # 3. Prepare parameters for the query

  ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
  x_geom <- sf_col_x %||% duckspatial:::get_geom_name(target_conn, x_list$query_name)
  if (length(x_geom != 0)) duckspatial:::assert_geometry_column(x_geom, x_list)

  ## 3.2. Get names of the rest of the columns
  x_rest <- duckspatial:::get_geom_name(
    target_conn,
    x_list$query_name,
    rest = TRUE,
    collapse = TRUE
  )

  ## 3.3. Build the base query
  x_geom <- if (length(x_geom) > 0) x_geom else "geometry"
  base.query <- glue::glue("
    SELECT
      {x_rest}
      ST_GeomFromWKB(
        h3_cell_to_boundary_wkb(
          h3_latlng_to_cell(
            ST_Y({x_geom}),
            ST_X({x_geom}),
            {resolution}
          )
        )
      ) as {x_geom}
    FROM
      {x_list$query_name};
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    duckspatial:::create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
   duckspatial::: ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = "duckspatial",
      crs    = "EPSG:4326",
      x_geom = x_geom
    )
  }

}





#' @rdname ddbh3_points_to
#' @export
ddbh3_points_to_h3 <- function(
    x,
    resolution,
    new_column = "h3string",
    h3_format = "string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {


  # 0. Validate inputs
  duckspatial:::assert_character_scalar(new_column, "new_column")
  duckspatial:::assert_character_scalar(h3_format, "h3_format")
  duckspatial:::assert_conn_character(conn, x)
  duckspatial:::assert_name(name)
  duckspatial:::assert_logic(overwrite, "overwrite")
  duckspatial:::assert_logic(quiet, "quiet")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")
  duckspatial:::assert_geom_type(x = x, conn = conn, geom = "POINT", multi = TRUE)


  # 1. Prepare inputs

  ## 1.1. Pre-extract geometry column name to overwrite, if present
  crs_x    <- duckspatial::ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")

  if (crs_x$input != "EPSG:4326") {
    cli::cli_abort("The CRS of the input must be {.val EPSG:4326}, not {.val crs_x$input}.")
  }

  ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df,
  ## validate character table names
  x <- suppressWarnings(duckspatial:::normalize_spatial_input(x, conn))

  ## 1.3. Resolve connections and handle imports
  resolve_conn <- duckspatial:::resolve_spatial_connections(x, y = NULL, conn = conn)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get query list of table names
  x_list <- duckspatial:::get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  ## 1.5. Install and load h3 on target connection
  # duckspatial::ddbs_install(target_conn, upgrade = FALSE, quiet = TRUE, extension = "h3")
  # duckspatial::ddbs_load(target_conn, quiet = TRUE, extension = "h3")


  # 2. Prepare the query

  ## 2.1. Get names of geometry columns (use saved sf_col_x from before transformation)
  x_geom <- sf_col_x %||% duckspatial:::get_geom_name(target_conn, x_list$query_name)
  duckspatial:::assert_geometry_column(x_geom, x_list)

  ## 2.2. Get names of the rest of the columns
  x_rest <- duckspatial:::get_geom_name(
    target_conn,
    x_list$query_name,
    rest = TRUE,
    collapse = TRUE
  )

  ## 2.3. Build the base query
  st_function <- switch(
    h3_format,
    "string" = glue::glue("h3_latlng_to_cell_string"),
    "bigint" = glue::glue("h3_latlng_to_cell"),
    cli::cli_abort("The {.arg h3_format} is not valid. Valid options: {.val {c('string', 'bigint')}}")
  )
  base.query <- glue::glue("
    SELECT
      * EXCLUDE {x_geom},
      {st_function}(
        ST_Y({x_geom}),
        ST_X({x_geom}),
        {resolution}
      ) as {new_column},
      {duckspatial:::build_geom_query(x_geom, name, crs_x, 'duckspatial')} AS {x_geom}
    FROM
      {x_list$query_name};
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    duckspatial:::create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
   duckspatial::: ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = "duckspatial",
      crs    = "EPSG:4326",
      x_geom = x_geom
    )
  }

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## create duckdb connection
conn_test <- duckh3::ddbh3_create_conn()

## Load example data
test_data <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
) |>
  duckspatial::ddbs_as_points()

# 1. points_to_spatial() ---------------------------------------------------------

## 1.1. Input data in different formats ----------

testthat::describe("ddbh3_points_to_spatial() works in different formats", {

  ## FORMAT 1 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_points_to_spatial(test_data_sf, resolution = 8)
    expect_s3_class(res, "duckspatial_df")
    ## Check geometry
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

  ## FORMAT 2 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_points_to_spatial(test_data_ddbs, resolution = 8)
    expect_s3_class(res, "duckspatial_df")
    ## Check geometry
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

  ## FORMAT 3 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf and store
    test_data_sf <- test_data |> duckspatial::ddbs_collect()
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_points_to_spatial("sf_pts", resolution = 8, conn = conn_test)
    ## Check class
    expect_s3_class(res, "duckspatial_df")
    ## Check geometry
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

})


testthat::test_that("returns one row per input row", {
  res <- ddbh3_points_to_spatial(test_data, resolution = 8) |>
    dplyr::collect()
  expect_equal(nrow(res), nrow(dplyr::collect(test_data)))
})

testthat::test_that("geometry column contains polygons", {
  res <- ddbh3_points_to_spatial(test_data, resolution = 8)
  expect_true(all(duckspatial::ddbs_geometry_type(res) == "POLYGON"))
})

testthat::test_that("different resolutions return different geometries", {
  res_fine <- ddbh3_points_to_spatial(test_data, resolution = 10) |>
    dplyr::collect()
  res_coarse <- ddbh3_points_to_spatial(test_data, resolution = 5) |>
    dplyr::collect()
  expect_false(identical(res_fine$geometry, res_coarse$geometry))
})

testthat::test_that("finer resolution returns smaller polygons", {
  res_fine <- ddbh3_points_to_spatial(test_data, resolution = 10) 
  res_coarse <- ddbh3_points_to_spatial(test_data, resolution = 5)
  expect_true(all(
    duckspatial::ddbs_area(res_fine, mode = "sf") < 
      duckspatial::ddbs_area(res_coarse, mode = "sf")
  ))
})

testthat::test_that("rejects non-EPSG:4326 input", {
  ## Project to a different CRS
  test_data_3857 <- test_data |>
    duckspatial::ddbs_transform(3857)
  expect_error(ddbh3_points_to_spatial(test_data_3857, resolution = 8))
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_points_to_spatial() arguments work", {

  ## ARGUMENT 1 - RESOLUTION
  testthat::it("resolution argument works", {
    ## Sample of valid resolutions return results
    for (r in c(0, 3, 7, 15)) {
      res <- ddbh3_points_to_spatial(test_data, resolution = r)
      expect_s3_class(res, "duckspatial_df")
      res_col <- dplyr::collect(res)
      expect_in("geometry", colnames(res_col))
    }
  })

  ## ARGUMENT 2 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    test_data_sf <- duckspatial::ddbs_collect(test_data)
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_points_to_spatial(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_points_to_spatial(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_points_to_spatial(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_points_to_spatial(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  ## Prepare sf input and connection
  conn_test <- ddbh3_create_conn()
  duckspatial::ddbs_write_table(conn_test, dplyr::collect(test_data), "sf_pts")

  ## Tests
  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_points_to_spatial(test_data, resolution = NULL))
    expect_error(ddbh3_points_to_spatial(test_data, resolution = "a"))
    expect_error(ddbh3_points_to_spatial(test_data, resolution = 1.5))
    expect_error(ddbh3_points_to_spatial(test_data, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_points_to_spatial(test_data, resolution = -1))
    expect_error(ddbh3_points_to_spatial(test_data, resolution = 16))
  })

  it("requires point geometry input", {
    ## Create a polygon input instead of points
    test_data_poly <- test_data |>
      ddbh3_lonlat_to_spatial(resolution = 8)
    expect_error(ddbh3_points_to_spatial(test_data_poly, resolution = 8))
  })

  it("requires EPSG:4326 CRS", {
    test_data_projected <- test_data |>
      duckspatial::ddbs_transform(3857)
    expect_error(ddbh3_points_to_spatial(test_data_projected, resolution = 8))
  })

  it("requires connection when using table names", {
    expect_error(ddbh3_points_to_spatial("sf_pts", resolution = 8, conn = NULL))
  })

  it("validates x argument type", {
    expect_error(ddbh3_points_to_spatial(x = 999, resolution = 8))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_points_to_spatial(test_data, resolution = 8, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_points_to_spatial(test_data, resolution = 8, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_points_to_spatial(test_data, resolution = 8, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_points_to_spatial("999", resolution = 8, conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_points_to_spatial(
      test_data,
      resolution = 8,
      conn = conn_test,
      name = c("banana", "banana")
    ))
  })

})



# 2. points_to_h3() ---------------------------------------------------------

## 2.1. Input data in different formats ----------

testthat::describe("ddbh3_points_to_h3() works in different formats", {

  ## FORMAT 1 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_points_to_h3(test_data_sf, resolution = 8)
    expect_s3_class(res, "duckspatial_df")
    ## Check column
    res_col <- dplyr::collect(res)
    expect_in("h3string", colnames(res_col))
    expect_type(res_col$h3string, "character")
  })

  ## FORMAT 2 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_points_to_h3(test_data_ddbs, resolution = 8)
    expect_s3_class(res, "duckspatial_df")
    ## Check column
    res_col <- dplyr::collect(res)
    expect_in("h3string", colnames(res_col))
    expect_type(res_col$h3string, "character")
  })

  ## FORMAT 3 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Store table in connection
    test_data_sf <- duckspatial::ddbs_collect(test_data)
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_points_to_h3("sf_pts", resolution = 8, conn = conn_test)
    ## Check class
    expect_s3_class(res, "duckspatial_df")
    ## Check column
    res_col <- dplyr::collect(res)
    expect_in("h3string", colnames(res_col))
    expect_type(res_col$h3string, "character")
  })

})


testthat::test_that("returns one row per input row", {
  res <- ddbh3_points_to_h3(test_data, resolution = 8) |>
    dplyr::collect()
  expect_equal(nrow(res), nrow(dplyr::collect(test_data)))
})

testthat::test_that("geometry column is preserved", {
  res <- ddbh3_points_to_h3(test_data, resolution = 8) |>
    dplyr::collect()
  expect_in("geometry", colnames(res))
})

testthat::test_that("string format returns valid h3 cells", {
  res <- ddbh3_points_to_h3(test_data, resolution = 8, h3_format = "string") |>
    dplyr::collect()
  validity <- ddbh3_is_h3(data.frame(h3string = res$h3string)) |>
    dplyr::collect()
  expect_true(all(validity$ish3))
})

testthat::test_that("bigint format returns integer64 column", {
  res <- ddbh3_points_to_h3(test_data, resolution = 8, h3_format = "bigint") |>
    dplyr::collect()
  expect_true(inherits(res$h3string, "integer64"))
})

testthat::test_that("returned cells have the requested resolution", {
  res <- ddbh3_points_to_h3(test_data, resolution = 5, h3_format = "string") |>
    dplyr::collect()
  res_resolution <- ddbh3_get_resolution(
    data.frame(h3string = res$h3string)
  ) |>
    dplyr::collect()
  expect_true(all(res_resolution$h3resolution == 5))
})

testthat::test_that("different resolutions return different cells", {
  res_fine <- ddbh3_points_to_h3(test_data, resolution = 10) |>
    dplyr::collect()
  res_coarse <- ddbh3_points_to_h3(test_data, resolution = 5) |>
    dplyr::collect()
  expect_false(identical(res_fine$h3string, res_coarse$h3string))
})

testthat::test_that("rejects non-EPSG:4326 input", {
  test_data_3857 <- test_data |>
    duckspatial::ddbs_transform(3857)
  expect_error(ddbh3_points_to_h3(test_data_3857, resolution = 8))
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_points_to_h3() arguments work", {

  ## ARGUMENT 1 - RESOLUTION
  testthat::it("resolution argument works", {
    for (r in c(0, 3, 7, 15)) {
      res <- ddbh3_points_to_h3(test_data, resolution = r)
      expect_s3_class(res, "duckspatial_df")
      res_col <- dplyr::collect(res)
      expect_type(res_col$h3string, "character")
    }
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_points_to_h3(test_data, resolution = 8, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 3 - H3_FORMAT
  testthat::it("h3_format argument works", {
    ## string format returns character
    res_string <- ddbh3_points_to_h3(test_data, resolution = 8, h3_format = "string")
    res_string_col <- dplyr::collect(res_string)
    expect_type(res_string_col$h3string, "character")
    ## bigint format returns integer64
    res_bigint <- ddbh3_points_to_h3(test_data, resolution = 8, h3_format = "bigint")
    res_bigint_col <- dplyr::collect(res_bigint)
    expect_true(inherits(res_bigint_col$h3string, "integer64"))
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    test_data_sf <- duckspatial::ddbs_collect(test_data)
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_points_to_h3(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_points_to_h3(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_points_to_h3(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_points_to_h3(
      test_data_sf,
      resolution = 8,
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  ## Prepare connection
  conn_test <- ddbh3_create_conn()
  duckspatial::ddbs_write_table(conn_test, dplyr::collect(test_data), "sf_pts")

  ## Tests
  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = NULL))
    expect_error(ddbh3_points_to_h3(test_data, resolution = "a"))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 1.5))
    expect_error(ddbh3_points_to_h3(test_data, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = -1))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 16))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, new_column = NULL))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, new_column = FALSE))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, new_column = 25))
  })

  it("requires h3_format to be a valid option", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, h3_format = "invalid"))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, h3_format = 999))
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, h3_format = NULL))
  })

  it("requires point geometry input", {
    test_data_poly <- test_data |>
      ddbh3_lonlat_to_spatial(resolution = 8)
    expect_error(ddbh3_points_to_h3(test_data_poly, resolution = 8))
  })

  it("requires EPSG:4326 CRS", {
    test_data_projected <- test_data |>
      duckspatial::ddbs_transform(3857)
    expect_error(ddbh3_points_to_h3(test_data_projected, resolution = 8))
  })

  it("requires connection when using table names", {
    expect_error(ddbh3_points_to_h3("sf_pts", resolution = 8, conn = NULL))
  })

  it("validates x argument type", {
    expect_error(ddbh3_points_to_h3(x = 999, resolution = 8))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_points_to_h3(test_data, resolution = 8, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_points_to_h3("999", resolution = 8, conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_points_to_h3(
      test_data,
      resolution = 8,
      conn = conn_test,
      name = c("banana", "banana")
    ))
  })

})

#' Get parent and children H3 cells
#'
#' Get the parent or children H3 cells of H3 cell indexes stored as strings or
#' unsigned 64-bit integers (`UBIGINT`) at a specified resolution:
#' `ddbh3_get_parent()`, `ddbh3_get_center_child()`, `ddbh3_get_children()`,
#' and `ddbh3_get_n_children()`.
#'
#'
#' @template x
#' @template resolution
#' @template h3
#' @template new_column
#' @template conn_null
#' @template name
#' @template nested
#' @template overwrite
#' @template quiet
#' 
#' @details
#' The four functions differ in the type of related cell they retrieve:
#' * `ddbh3_get_parent()` returns the parent cell at a coarser resolution
#' * `ddbh3_get_center_child()` returns the center child cell at a finer resolution
#' * `ddbh3_get_children()` returns all children cells at a finer resolution
#' * `ddbh3_get_n_children()` returns the number of children cells at a finer
#'   resolution, without computing them
#'
#' @template returns_tbl
#'
#' @name ddbh3_get_hierarchy
#' @aliases ddbh3_get_parent ddbh3_get_center_child ddbh3_get_children ddbh3_get_n_children
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add h3 strings
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)
#' 
#' ## GET PARENTS ----------
#' 
#' ## Get resolution-7 parent
#' points_parent_tbl <- ddbh3_get_parent(points_tbl, resolution = 7)
#' 
#' ## Check the resolution
#' ddbh3_get_resolution(
#'   points_parent_tbl,
#'   h3 = "h3parent"
#' )
#' 
#' ## Add with mutate
#' points_tbl |> 
#'   mutate(parent4 = ddbh3_get_parent(h3string, 4))
#' 
#' ## GET CHILDREN ----------
#' 
#' ## Get level 9 children
#' children_9_tbl <- ddbh3_get_children(points_tbl, resolution = 9)
#' 
#' ## Get level 9 children in a nested list
#' children_9_nested_tbl <- ddbh3_get_children(points_tbl, resolution = 9, nested = TRUE)
#' 
#' ## Add with mutate (nested)
#' points_tbl |> 
#'   mutate(children9 = ddbh3_get_children(h3string, 9))
#' 
#' ## Add with mutate (unnested)
#' points_tbl |> 
#'   mutate(children9 = ddbh3_get_children(h3string, 9)) |> 
#'   mutate(children9 = unnest(children9))
#' 
#' ## GET CENTER CHILD ------
#' 
#' ## Get the center child of res 10 (1 child per row)
#' center_child_10_tbl <- ddbh3_get_center_child(points_tbl, resolution = 10)
#' 
#' ## Add with mutate
#' points_tbl |> 
#'   mutate(center = ddbh3_get_center_child(h3string, 9))
#' 
#' ## NUMBER OF CHILDREN -----
#' 
#' ## How many children of level 10 does each level 8 have?
#' n_children_tbl <- ddbh3_get_n_children(points_tbl, resolution = 10)
#' 
#' ## Add with mutate
#' points_tbl |> 
#'   mutate(n_children = ddbh3_get_n_children(h3string, 15))
NULL




#' @rdname ddbh3_get_hierarchy
#' @export
ddbh3_get_parent <- function(
    x,
    resolution = 8,
    h3 = "h3string",
    new_column = "h3parent",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_parent({h3}, {resolution})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_parent(x, {resolution})")
  ) 

}





#' @rdname ddbh3_get_hierarchy
#' @export
ddbh3_get_children <- function(
    x,
    resolution = 8,
    h3 = "h3string",
    new_column = "h3children",
    conn = NULL,
    name = NULL,
    nested = FALSE,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_logic(nested, "nested")
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  if (isTRUE(nested)) {
    built_fun <- glue::glue("h3_cell_to_children({h3}, {resolution})")
  } else {
    built_fun <- glue::glue("UNNEST(h3_cell_to_children({h3}, {resolution}))")
  }

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_children(x, {resolution})")
  ) 

}





#' @rdname ddbh3_get_hierarchy
#' @export
ddbh3_get_n_children <- function(
    x,
    resolution = 8,
    h3 = "h3string",
    new_column = "h3n_children",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_children_size({h3}, {resolution})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_children_size(x, {resolution})")
  ) 

}





#' @rdname ddbh3_get_hierarchy
#' @export
ddbh3_get_center_child <- function(
    x,
    resolution = 8,
    h3 = "h3string",
    new_column = "h3center_child",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_center_child({h3}, {resolution})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_center_child(x, {resolution})")
  ) 

}






#' Get the icosahedron faces of H3 cell indexes
#'
#' Get the icosahedron faces intersected by H3 cell indexes stored as strings
#' or unsigned 64-bit integers (`UBIGINT`). Each H3 cell maps onto one or more
#' of the 20 faces of the underlying icosahedron used to construct the H3 grid.
#'
#' @template x
#' @template h3
#' @template new_column
#' @template conn_null
#' @template name
#' @template nested
#' @template overwrite
#' @template quiet
#'
#' @template returns_tbl
#' @export
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add H3 string column
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 6)
#' 
#' ## Get faces (unnested)
#' faces_tbl <- ddbh3_get_icosahedron_faces(points_tbl)
#' 
#' ## Get faces (nested)
#' faces_nested_tbl <- ddbh3_get_icosahedron_faces(points_tbl, nested = TRUE)
#' 
#' ## Add using mutate (nested)
#' points_tbl |> 
#'   mutate(faces = ddbh3_get_icosahedron_faces(h3string))
#' 
#' ## Add using mutate (unnested)
#' points_tbl |> 
#'   mutate(faces = ddbh3_get_icosahedron_faces(h3string)) |> 
#'   mutate(faces_unnested = unnest(faces))
ddbh3_get_icosahedron_faces <- function(
    x,
    h3 = "h3string",
    new_column = "h3faces",
    conn = NULL,
    name = NULL,
    nested = FALSE,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_logic(nested, "nested")

  # 1. Build parameters string  
  if (isTRUE(nested)) {
    built_fun <- glue::glue("h3_get_icosahedron_faces({h3})")
  } else {
    built_fun <- glue::glue("UNNEST(h3_get_icosahedron_faces({h3}))")
  }

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_get_icosahedron_faces(x)")
  ) 

}





#' Get the position of an H3 cell within its parent
#'
#' Get the position of H3 cell indexes stored as strings or unsigned 64-bit
#' integers (`UBIGINT`) within their parent cell at a specified resolution.
#' The position is a zero-based index among all children of the parent cell.
#'
#' @template x
#' @template h3
#' @template resolution
#' @template new_column
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#'
#' @template returns_tbl
#' @export
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add H3 string column
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 6)
#' 
#' ## Get position relative to resolution 4
#' ddbh3_get_child_pos(points_tbl, resolution = 4)
ddbh3_get_child_pos <- function(
    x,
    resolution = 8,
    h3 = "h3string",
    new_column = "h3child_pos",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_character_scalar(new_column, "new_column")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")


  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_child_pos({h3}, {resolution})")


  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_child_pos(x, {resolution})")
  ) 

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## create duckdb connection
conn_test <- duckh3::ddbh3_create_conn()

## Load example data
test_data <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
) |> 
  ddbh3_lonlat_to_h3(resolution = 10)

test_data_5 <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
) |> 
  head(1) |> 
  ddbh3_lonlat_to_h3(resolution = 5)

# 1. get_parent() ---------------------------------------------------------

## 1.1. Input data in different formats ----------

testthat::describe("ddbh3_get_parent() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_parent(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_get_parent(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_parent(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_parent(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_parent("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

})


testthat::test_that("parent has lower resolution than child", {
  res <- ddbh3_get_parent(test_data, resolution = 5) |>
    dplyr::collect()
  ## Get resolution of parent cells
  parent_res <- ddbh3_get_resolution(
    data.frame(h3string = res$h3parent)
  ) |>
    dplyr::collect()
  expect_true(all(parent_res$h3resolution == 5))
})

testthat::test_that("lower resolution returns coarser parent", {
  res_coarse <- ddbh3_get_parent(test_data, resolution = 3) |>
    dplyr::collect()
  res_fine <- ddbh3_get_parent(test_data, resolution = 6) |>
    dplyr::collect()
  ## Coarser parent strings should differ from finer parent strings
  expect_false(identical(res_coarse$h3parent, res_fine$h3parent))
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_get_parent() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_parent(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3parent", colnames(res_col))
    expect_type(res_col$h3parent, "character")
  })

  ## ARGUMENT 2 - RESOLUTION
  testthat::it("resolution argument works", {
    ## All valid resolutions return results
    for (r in c(0, 3, 7, 15)) {
      res <- ddbh3_get_parent(test_data, resolution = r)
      res_col <- dplyr::collect(res)
      expect_type(res_col$h3parent, "character")
    }
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_parent(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_parent(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_parent(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_parent(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_parent(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_parent(test_data, h3 = NULL))
    expect_error(ddbh3_get_parent(test_data, h3 = TRUE))
    expect_error(ddbh3_get_parent(test_data, h3 = 2))
  })

  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_get_parent(test_data, resolution = NULL))
    expect_error(ddbh3_get_parent(test_data, resolution = "a"))
    expect_error(ddbh3_get_parent(test_data, resolution = 1.5))
    expect_error(ddbh3_get_parent(test_data, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_get_parent(test_data, resolution = -1))
    expect_error(ddbh3_get_parent(test_data, resolution = 16))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_parent(test_data, new_column = NULL))
    expect_error(ddbh3_get_parent(test_data, new_column = FALSE))
    expect_error(ddbh3_get_parent(test_data, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_get_parent("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_parent(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_parent(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_parent(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_parent(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_parent(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_parent(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 2. get_children() ---------------------------------------------------------

## 2.1. Input data in different formats ----------

testthat::describe("ddbh3_get_children() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_children(test_data_5)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_5_df <- dplyr::collect(test_data_5)
    res <- ddbh3_get_children(test_data_5_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_5_sf <- test_data_5 |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_children(test_data_5_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_5_ddbs <- test_data_5 |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_children(test_data_5_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_5_sf <- test_data_5 |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_5_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_children("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

})


testthat::test_that("unnested result has more rows than nested", {
  ## Unnested (default): one row per child
  res_unnested <- ddbh3_get_children(test_data_5, nested = FALSE) |>
    dplyr::collect()
  ## Nested: one row per parent cell, children column contains a list
  res_nested <- ddbh3_get_children(test_data_5, nested = TRUE) |>
    dplyr::collect()
  expect_gt(nrow(res_unnested), nrow(res_nested))
})

testthat::test_that("nested result has same number of rows as input", {
  res_nested <- ddbh3_get_children(test_data_5, nested = TRUE) |>
    dplyr::collect()
  expect_equal(nrow(res_nested), nrow(dplyr::collect(test_data_5)))
})

testthat::test_that("children have higher resolution than parent", {
  res <- ddbh3_get_children(test_data_5, resolution = 6, nested = FALSE) |>
    dplyr::collect()
  ## Get resolution of children cells
  children_res <- ddbh3_get_resolution(
    data.frame(h3string = res$h3children)
  ) |>
    dplyr::collect()
  expect_true(all(children_res$h3resolution == 6))
})

testthat::test_that("higher resolution returns more children", {
  res_fine <- ddbh3_get_children(test_data_5, resolution = 7, nested = FALSE) |>
    dplyr::collect()
  res_coarse <- ddbh3_get_children(test_data_5, resolution = 6, nested = FALSE) |>
    dplyr::collect()
  expect_gt(nrow(res_fine), nrow(res_coarse))
})


## 2.2. Arguments work ------------

testthat::describe("ddbh3_get_children() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_5_mod <- test_data_5 |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_children(test_data_5_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3children", colnames(res_col))
    expect_type(res_col$h3children, "character")
  })

  ## ARGUMENT 2 - RESOLUTION
  testthat::it("resolution argument works", {
    ## Sample of valid resolutions return results
    for (r in c(7, 8, 9)) {
      res <- ddbh3_get_children(test_data_5, resolution = r)
      res_col <- dplyr::collect(res)
      expect_type(res_col$h3children, "character")
    }
  })

  ## ARGUMENT 3 - NESTED
  testthat::it("nested argument works", {
    ## nested = FALSE returns character column (unnested)
    res_false <- ddbh3_get_children(test_data_5, nested = FALSE)
    res_false_col <- dplyr::collect(res_false)
    expect_type(res_false_col$h3children, "character")
    ## nested = TRUE returns list column
    res_true <- ddbh3_get_children(test_data_5, nested = TRUE)
    res_true_col <- dplyr::collect(res_true)
    expect_type(res_true_col$h3children, "list")
  })

  ## ARGUMENT 4 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_children(test_data_5, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 5 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl",
      overwrite = TRUE
    ))
  })

})


## 2.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_5_tbl", dplyr::collect(test_data_5))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_children(test_data_5, h3 = NULL))
    expect_error(ddbh3_get_children(test_data_5, h3 = TRUE))
    expect_error(ddbh3_get_children(test_data_5, h3 = 2))
  })

  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_get_children(test_data_5, resolution = NULL))
    expect_error(ddbh3_get_children(test_data_5, resolution = "a"))
    expect_error(ddbh3_get_children(test_data_5, resolution = 1.5))
    expect_error(ddbh3_get_children(test_data_5, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_get_children(test_data_5, resolution = -1))
    expect_error(ddbh3_get_children(test_data_5, resolution = 16))
  })

  it("requires nested argument as logical", {
    expect_error(ddbh3_get_children(test_data_5, nested = NULL))
    expect_error(ddbh3_get_children(test_data_5, nested = "yes"))
    expect_error(ddbh3_get_children(test_data_5, nested = 1))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_children(test_data_5, new_column = NULL))
    expect_error(ddbh3_get_children(test_data_5, new_column = FALSE))
    expect_error(ddbh3_get_children(test_data_5, new_column = 25))
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_children(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_children(test_data_5, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_children(test_data_5, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_children(test_data_5, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_children(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_children(test_data_5, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 3. get_n_children() ---------------------------------------------------------

## 3.1. Input data in different formats ----------

testthat::describe("ddbh3_get_n_children() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_n_children(test_data_5)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_5_df <- dplyr::collect(test_data_5)
    res <- ddbh3_get_n_children(test_data_5_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_5_sf <- test_data_5 |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_n_children(test_data_5_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_5_ddbs <- test_data_5 |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_n_children(test_data_5_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_5_sf <- test_data_5 |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_5_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_n_children("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

})


testthat::test_that("n_children matches actual number of children", {
  ## Get number of children reported by the function
  res_n <- ddbh3_get_n_children(test_data_5, resolution = 10) |>
    dplyr::collect()
  ## Get actual number of children by unnesting
  res_children <- ddbh3_get_children(test_data_5, resolution = 10, nested = FALSE) |>
    dplyr::collect()
  expect_equal(sum(res_n$h3n_children), nrow(res_children))
})

testthat::test_that("n_children is positive", {
  res <- ddbh3_get_n_children(test_data_5, resolution = 10) |>
    dplyr::collect()
  expect_true(all(res$h3n_children > 0))
})

testthat::test_that("finer resolution returns more children", {
  res_fine <- ddbh3_get_n_children(test_data_5, resolution = 11) |>
    dplyr::collect()
  res_coarse <- ddbh3_get_n_children(test_data_5, resolution = 9) |>
    dplyr::collect()
  expect_true(all(res_fine$h3n_children > res_coarse$h3n_children))
})


## 3.2. Arguments work ------------

testthat::describe("ddbh3_get_n_children() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_5_mod <- test_data_5 |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_n_children(test_data_5_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3n_children", colnames(res_col))
    expect_s3_class(res_col$h3n_children, "integer64")
  })

  ## ARGUMENT 2 - RESOLUTION
  testthat::it("resolution argument works", {
    ## Sample of valid resolutions return results
    for (r in c(9, 11, 13, 15)) {
      res <- ddbh3_get_n_children(test_data_5, resolution = r)
      res_col <- dplyr::collect(res)
      expect_s3_class(res_col$h3n_children, "integer64")
    }
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_n_children(test_data_5, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_s3_class(res_col$res, "integer64")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_n_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_n_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_n_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_n_children(
      dplyr::collect(test_data_5),
      new_column = "res",
      conn = conn_test,
      name = "test_data_5_tbl",
      overwrite = TRUE
    ))
  })

})


## 3.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_5_tbl", dplyr::collect(test_data_5))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_n_children(test_data_5, h3 = NULL))
    expect_error(ddbh3_get_n_children(test_data_5, h3 = TRUE))
    expect_error(ddbh3_get_n_children(test_data_5, h3 = 2))
  })

  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_get_n_children(test_data_5, resolution = NULL))
    expect_error(ddbh3_get_n_children(test_data_5, resolution = "a"))
    expect_error(ddbh3_get_n_children(test_data_5, resolution = 1.5))
    expect_error(ddbh3_get_n_children(test_data_5, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_get_n_children(test_data_5, resolution = -1))
    expect_error(ddbh3_get_n_children(test_data_5, resolution = 16))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_n_children(test_data_5, new_column = NULL))
    expect_error(ddbh3_get_n_children(test_data_5, new_column = FALSE))
    expect_error(ddbh3_get_n_children(test_data_5, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_get_n_children("test_data_5_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_n_children(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_n_children(test_data_5, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_n_children(test_data_5, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_n_children(test_data_5, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_n_children(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_n_children(test_data_5, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 4. get_center_child() ---------------------------------------------------------

## 4.1. Input data in different formats ----------

testthat::describe("ddbh3_get_center_child() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_center_child(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_get_center_child(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_center_child(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_center_child(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_center_child("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

})


testthat::test_that("center child has the requested resolution", {
  res <- ddbh3_get_center_child(test_data, resolution = 10) |>
    dplyr::collect()
  ## Verify resolution of center child cells
  child_res <- ddbh3_get_resolution(
    data.frame(h3string = res$h3center_child)
  ) |>
    dplyr::collect()
  expect_true(all(child_res$h3resolution == 10))
})

testthat::test_that("center child is among the children", {
  ## Get center child
  res_center <- ddbh3_get_center_child(test_data, resolution = 10) |>
    dplyr::collect()
  ## Get all children
  res_children <- ddbh3_get_children(test_data, resolution = 10, nested = FALSE) |>
    dplyr::collect()
  ## Center child must be a subset of all children
  expect_true(all(res_center$h3center_child %in% res_children$h3children))
})

testthat::test_that("returns one row per input row", {
  res <- ddbh3_get_center_child(test_data, resolution = 10) |>
    dplyr::collect()
  expect_equal(nrow(res), nrow(dplyr::collect(test_data)))
})

testthat::test_that("different resolutions return different center children", {
  res_fine <- ddbh3_get_center_child(test_data, resolution = 11) |>
    dplyr::collect()
  res_coarse <- ddbh3_get_center_child(test_data, resolution = 9) |>
    dplyr::collect()
  expect_false(identical(res_fine$h3center_child, res_coarse$h3center_child))
})


## 4.2. Arguments work ------------

testthat::describe("ddbh3_get_center_child() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_center_child(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3center_child", colnames(res_col))
    expect_type(res_col$h3center_child, "character")
  })

  ## ARGUMENT 2 - RESOLUTION
  testthat::it("resolution argument works", {
    ## Sample of valid resolutions return results
    for (r in c(9, 11, 13, 15)) {
      res <- ddbh3_get_center_child(test_data, resolution = r)
      res_col <- dplyr::collect(res)
      expect_type(res_col$h3center_child, "character")
    }
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_center_child(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_center_child(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_center_child(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_center_child(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_center_child(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 4.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_center_child(test_data, h3 = NULL))
    expect_error(ddbh3_get_center_child(test_data, h3 = TRUE))
    expect_error(ddbh3_get_center_child(test_data, h3 = 2))
  })

  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_get_center_child(test_data, resolution = NULL))
    expect_error(ddbh3_get_center_child(test_data, resolution = "a"))
    expect_error(ddbh3_get_center_child(test_data, resolution = 1.5))
    expect_error(ddbh3_get_center_child(test_data, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_get_center_child(test_data, resolution = -1))
    expect_error(ddbh3_get_center_child(test_data, resolution = 16))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_center_child(test_data, new_column = NULL))
    expect_error(ddbh3_get_center_child(test_data, new_column = FALSE))
    expect_error(ddbh3_get_center_child(test_data, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_get_center_child("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_center_child(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_center_child(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_center_child(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_center_child(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_center_child(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_center_child(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 5. get_icosahedron_faces() ---------------------------------------------------------

## 5.1. Input data in different formats ----------

testthat::describe("ddbh3_get_icosahedron_faces() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_icosahedron_faces(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_get_icosahedron_faces(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_icosahedron_faces(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_icosahedron_faces(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_icosahedron_faces("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

})


# testthat::test_that("unnested result has more rows than nested", {
#   ## Unnested (default): one row per face
#   res_unnested <- ddbh3_get_icosahedron_faces(test_data, nested = FALSE) |>
#     dplyr::collect()
#   ## Nested: one row per cell, faces column contains a list
#   res_nested <- ddbh3_get_icosahedron_faces(test_data, nested = TRUE) |>
#     dplyr::collect()
#   expect_gt(nrow(res_unnested), nrow(res_nested))
# })

testthat::test_that("nested result has same number of rows as input", {
  res_nested <- ddbh3_get_icosahedron_faces(test_data, nested = TRUE) |>
    dplyr::collect()
  expect_equal(nrow(res_nested), nrow(dplyr::collect(test_data)))
})

testthat::test_that("face indices are within valid range", {
  ## Icosahedron has 20 faces, indexed 0-19
  res <- ddbh3_get_icosahedron_faces(test_data, nested = FALSE) |>
    dplyr::collect()
  expect_true(all(res$h3faces >= 0 & res$h3faces <= 19))
})


## 5.2. Arguments work ------------

testthat::describe("ddbh3_get_icosahedron_faces() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_icosahedron_faces(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3faces", colnames(res_col))
    expect_type(res_col$h3faces, "integer")
  })

  ## ARGUMENT 2 - NESTED
  testthat::it("nested argument works", {
    ## nested = FALSE returns integer column (unnested)
    res_false <- ddbh3_get_icosahedron_faces(test_data, nested = FALSE)
    res_false_col <- dplyr::collect(res_false)
    expect_type(res_false_col$h3faces, "integer")
    ## nested = TRUE returns list column
    res_true <- ddbh3_get_icosahedron_faces(test_data, nested = TRUE)
    res_true_col <- dplyr::collect(res_true)
    expect_type(res_true_col$h3faces, "list")
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_icosahedron_faces(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "integer")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_icosahedron_faces(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_icosahedron_faces(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_icosahedron_faces(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_icosahedron_faces(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 5.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, h3 = NULL))
    expect_error(ddbh3_get_icosahedron_faces(test_data, h3 = TRUE))
    expect_error(ddbh3_get_icosahedron_faces(test_data, h3 = 2))
  })

  it("requires nested argument as logical", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, nested = NULL))
    expect_error(ddbh3_get_icosahedron_faces(test_data, nested = "yes"))
    expect_error(ddbh3_get_icosahedron_faces(test_data, nested = 1))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, new_column = NULL))
    expect_error(ddbh3_get_icosahedron_faces(test_data, new_column = FALSE))
    expect_error(ddbh3_get_icosahedron_faces(test_data, new_column = 25))
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_icosahedron_faces(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_icosahedron_faces(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_icosahedron_faces(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 6. get_child_pos() ---------------------------------------------------------

## 6.1. Input data in different formats ----------

testthat::describe("ddbh3_get_child_pos() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_get_child_pos(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_get_child_pos(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_get_child_pos(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_get_child_pos(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_get_child_pos("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

})


testthat::test_that("child position is non-negative", {
  res <- ddbh3_get_child_pos(test_data, resolution = 10) |>
    dplyr::collect()
  expect_true(all(res$h3child_pos >= 0))
})

testthat::test_that("child position is within the number of children at that resolution", {
  resolution <- 10
  ## Get child positions
  res_pos <- ddbh3_get_child_pos(test_data, resolution = resolution) |>
    dplyr::collect()
  ## Get number of children at that resolution
  res_n <- ddbh3_get_n_children(test_data, resolution = resolution) |>
    dplyr::collect()
  ## Position must be strictly less than the number of children
  expect_true(all(res_pos$h3child_pos < res_n$h3n_children))
})

testthat::test_that("returns one row per input row", {
  res <- ddbh3_get_child_pos(test_data, resolution = 10) |>
    dplyr::collect()
  expect_equal(nrow(res), nrow(dplyr::collect(test_data)))
})

testthat::test_that("different resolutions return different positions", {
  res_fine <- ddbh3_get_child_pos(test_data, resolution = 11) |>
    dplyr::collect()
  res_coarse <- ddbh3_get_child_pos(test_data, resolution = 9) |>
    dplyr::collect()
  expect_false(identical(res_fine$h3child_pos, res_coarse$h3child_pos))
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_get_child_pos() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_get_child_pos(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3child_pos", colnames(res_col))
    expect_s3_class(res_col$h3child_pos, "integer64")
  })

  ## ARGUMENT 2 - RESOLUTION
  testthat::it("resolution argument works", {
    ## Sample of valid resolutions return results
    for (r in c(9, 11, 13, 15)) {
      res <- ddbh3_get_child_pos(test_data, resolution = r)
      res_col <- dplyr::collect(res)
      expect_s3_class(res_col$h3child_pos, "integer64")
    }
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_get_child_pos(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_s3_class(res_col$res, "integer64")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_get_child_pos(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_get_child_pos(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_get_child_pos(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_get_child_pos(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_get_child_pos(test_data, h3 = NULL))
    expect_error(ddbh3_get_child_pos(test_data, h3 = TRUE))
    expect_error(ddbh3_get_child_pos(test_data, h3 = 2))
  })

  it("requires resolution to be an integer scalar", {
    expect_error(ddbh3_get_child_pos(test_data, resolution = NULL))
    expect_error(ddbh3_get_child_pos(test_data, resolution = "a"))
    expect_error(ddbh3_get_child_pos(test_data, resolution = 1.5))
    expect_error(ddbh3_get_child_pos(test_data, resolution = c(1, 2)))
  })

  it("requires resolution to be in the range 0-15", {
    expect_error(ddbh3_get_child_pos(test_data, resolution = -1))
    expect_error(ddbh3_get_child_pos(test_data, resolution = 16))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_get_child_pos(test_data, new_column = NULL))
    expect_error(ddbh3_get_child_pos(test_data, new_column = FALSE))
    expect_error(ddbh3_get_child_pos(test_data, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_get_child_pos("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_get_child_pos(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_get_child_pos(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_get_child_pos(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_get_child_pos(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_get_child_pos(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_get_child_pos(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})

#' @returns One of the following, depending on the inputs:
#'   \describe{
#'     \item{`tbl_lazy`}{If `x` is not spatial.}
#'     \item{`duckspatial_df`}{If `x` is spatial (e.g. an `sf` or `duckspatial_df` object).}
#'     \item{`TRUE` (invisibly)}{If `name` is provided, a table is created in the connection
#'       and `TRUE` is returned invisibly.}
#'     \item{vector}{If `x` is a character vector and `conn = NULL`, the function operates
#'       in vectorized mode, returning a vector of the same length as `x`.}
#'   }

create_ddbh3_macros <- function(conn) {

  macros <- list(

    # ── lonlat to --------------------------------------------------------------
    "CREATE OR REPLACE MACRO ddbh3_lonlat_to_h3(lon, lat, resolution) AS
      h3_latlng_to_cell_string(lat, lon, CAST(resolution AS INTEGER))",

    # ── h3 to  -----------------------------------------------------------------
    "CREATE OR REPLACE MACRO ddbh3_bigint_to_strings(h3) AS
      h3_h3_to_string(h3)",

    "CREATE OR REPLACE MACRO ddbh3_h3_to_lat(h3) AS
      h3_cell_to_lat(h3)",

    "CREATE OR REPLACE MACRO ddbh3_h3_to_lon(h3) AS
      h3_cell_to_lng(h3)",

    "CREATE OR REPLACE MACRO ddbh3_strings_to_bigint(h3) AS
      h3_string_to_h3(h3)",

    # ── H3 properties ----------------------------------------------------------
    "CREATE OR REPLACE MACRO ddbh3_get_resolution(h3) AS
      h3_get_resolution(h3)",

    "CREATE OR REPLACE MACRO ddbh3_is_pentagon(h3) AS
      h3_is_pentagon(h3)",

    "CREATE OR REPLACE MACRO ddbh3_is_res_class_iii(h3) AS
      h3_is_res_class_iii(h3)",

    "CREATE OR REPLACE MACRO ddbh3_is_h3(h3) AS
      h3_is_valid_cell(h3)",

    "CREATE OR REPLACE MACRO ddbh3_is_vertex(h3vertex) AS
      h3_is_valid_vertex(h3vertex)",

    # ── H3 hierarchy -----------------------------------------------------------
    "CREATE OR REPLACE MACRO ddbh3_get_parent(h3, resolution) AS
      h3_cell_to_parent(h3, CAST(resolution AS INTEGER))",

    "CREATE OR REPLACE MACRO ddbh3_get_children(h3, resolution) AS
      h3_cell_to_children(h3, CAST(resolution AS INTEGER))",

    "CREATE OR REPLACE MACRO ddbh3_get_center_child(h3, resolution) AS
      h3_cell_to_center_child(h3, CAST(resolution AS INTEGER))",

    "CREATE OR REPLACE MACRO ddbh3_get_n_children(h3, resolution) AS
      h3_cell_to_children_size(h3, CAST(resolution AS INTEGER))",

    "CREATE OR REPLACE MACRO ddbh3_get_child_pos(h3, resolution) AS
      h3_cell_to_child_pos(h3, CAST(resolution AS INTEGER))",
      
    "CREATE OR REPLACE MACRO ddbh3_get_icosahedron_faces(h3) AS
      h3_get_icosahedron_faces(h3)",

    # ── H3 vertices -------------------------------------------------------------
    "CREATE OR REPLACE MACRO ddbh3_h3_to_vertex(h3, n) AS
      h3_cell_to_vertex(h3, CAST(n AS INTEGER))",

    "CREATE OR REPLACE MACRO ddbh3_vertex_to_lat(h3vertex) AS
      h3_vertex_to_lat(h3vertex)",

    "CREATE OR REPLACE MACRO ddbh3_vertex_to_lon(h3vertex) AS
      h3_vertex_to_lng(h3vertex)"

  )

  invisible(lapply(macros, DBI::dbExecute, conn = conn))

}


create_ddbh3_default_macros <- function() {
  create_ddbh3_macros(duckspatial:::ddbs_default_conn())
}



check_nested_column <- function(x, column) {

  ## For lazy tables we query the column type. However, the way of
  ## finding the source conn and table is different depending of the
  ## object type
  if (inherits(x, "duckspatial_df") | inherits(x, "tbl_duckdb_connection")) {
    if (inherits(x, "duckspatial_df")) {
      conn     <- attr(x, "source_conn")
      tbl_name <- attr(x, "source_table")
    } else {
      conn <- dbplyr::remote_con(x)
      tbl_name <- dbplyr::remote_name(x)

      ## If the name is null, we need to compute the previous lazy
      ## table to obtain a table name to query
      if (is.null(tbl_name)) {
        tbl_comp <- dplyr::compute(x)
        tbl_name <- dbplyr::remote_name(tbl_comp)
      }
    }

    col_type <- DBI::dbGetQuery(
      conn,
      glue::glue("SELECT data_type FROM information_schema.columns 
                  WHERE table_name = '{tbl_name}' 
                  AND column_name = '{column}'")
    )$data_type

    return(grepl("\\[\\]", col_type))

    ## For data.frame and sf we look for a list of characters
  } else if (inherits(x, "data.frame") | inherits(x, "sf")) {
    
    if (inherits(x[[column]], "list")) return(TRUE) else return(FALSE)

    ## For every other case, we consider unnested and let downstream
    ## processes to manage it
  } else {
    return(FALSE)
  }
  
}


## e.g. ddbh3_h3_to_lat("8ad02dcc1947fff") 
get_vectorized_result <- function(x, fun) {

  ## Create a random view name
  conn      <- duckspatial:::ddbs_default_conn()
  view_name <- duckspatial:::ddbs_temp_view_name()

  ## Register the data as a view
  duckdb::duckdb_register(conn, view_name, data.frame(x = x))
  on.exit(duckdb::duckdb_unregister(conn, view_name))

  ## Apply function, and return the pulled vector
  res <- dplyr::tbl(conn, view_name) |> 
    dplyr::mutate(res = dbplyr::sql(glue::glue("{fun}"))) |> 
    dplyr::pull("res")

  if (any(is.na(res))) {
    cli::cli_warn("Some elements of {.arg x} are invalid. Returning NA.")
  }

  return(res)

}

#' Check properties of H3 cell indexes
#'
#' Check properties of H3 cell indexes stored as strings or unsigned 64-bit
#' integers
#'
#' @template x
#' @template h3
#' @template h3vertex
#' @template new_column
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#' 
#' @details
#' The functions check different properties of H3 cell indexes, all
#' returning a logical column:
#' * `ddbh3_is_pentagon()`: returns `TRUE` if the H3 cell is one of the 12
#'   pentagonal cells that exist at each H3 resolution
#' * `ddbh3_is_h3()`: returns `TRUE` if the H3 cell index is a valid H3 cell
#' * `ddbh3_is_res_class_iii()`: returns `TRUE` if the H3 cell belongs to a
#'   Class III resolution (odd resolutions: 1, 3, 5, 7, 9, 11, 13, 15)
#' * `ddbh3_is_vertex()`: returns `TRUE` if the index is a valid H3 vertex
#'
#' @template returns_tbl
#'
#' @name ddbh3_is
#' @aliases ddbh3_is_pentagon ddbh3_is_valid ddbh3_is_res_class_iii
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add h3 strings
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)
#' 
#' ## IS VALID H3 -------------
#' 
#' ## Check if h3 indexes are valid
#' ddbh3_is_h3(points_tbl)
#' 
#' ## Check in mutate
#' points_tbl |>
#'   mutate(valid = ddbh3_is_h3(h3string))
#' 
#' ## IS PENTAGON -------------
#' 
#' ## Check if h3 indexes are pentagons
#' ddbh3_is_pentagon(points_tbl)
#' 
#' ## Check in mutate
#' points_tbl |>
#'   mutate(is_pent = ddbh3_is_pentagon(h3string))
#' 
#' ## IS CLASS III ------------
#' 
#' ## Check if h3 indexes belong to a Class III resolution
#' ddbh3_is_res_class_iii(points_tbl)
#' 
#' ## Check across multiple resolutions
#' ddbh3_lonlat_to_h3(points_tbl, resolution = 7) |>
#'   ddbh3_is_res_class_iii()
#' 
#' ## IS VERTEX ---------------
#' 
#' ## Get vertexes first
#' vertex_tbl <- ddbh3_h3_to_vertex(points_tbl, n = 1)
#' 
#' ## Check if indexes are valid vertexes
#' ddbh3_is_vertex(vertex_tbl, h3 = "h3vertex")
#' 
#' ## Check in mutate (mix of h3 cells and vertexes)
#' vertex_tbl |>
#'   mutate(
#'     cell_valid  = ddbh3_is_h3(h3string),
#'     vertex_valid = ddbh3_is_vertex(h3vertex)
#'   )
NULL




#' @rdname ddbh3_is
#' @export
ddbh3_is_pentagon <- function(
    x,
    h3 = "h3string",
    new_column = "ispentagon",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_is_pentagon({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_is_pentagon(x)"
  ) 

}




#' @rdname ddbh3_is
#' @export
ddbh3_is_h3 <- function(
    x,
    h3 = "h3string",
    new_column = "ish3",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_is_valid_cell({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_is_valid_cell(x)"
  ) 

}





#' @rdname ddbh3_is
#' @export
ddbh3_is_res_class_iii <- function(
    x,
    h3 = "h3string",
    new_column = "isclassiii",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_is_res_class_iii({h3})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_is_res_class_iii(x)"
  ) 

}




#' @rdname ddbh3_is
#' @export
ddbh3_is_vertex <- function(
    x,
    h3vertex = "h3vertex",
    new_column = "isvertex",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3vertex, "h3vertex")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_is_valid_vertex({h3vertex})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_is_valid_vertex(x)"
  ) 

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## create duckdb connection
conn_test <- duckh3::ddbh3_create_conn()

## Load example data
test_data <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
) |> 
  ddbh3_lonlat_to_h3(resolution = 10)


# 1. is_pentagon() ---------------------------------------------------------

## 1.1. Input data in different formats ----------

testthat::describe("ddbh3_is_pentagon() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_is_pentagon(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_is_pentagon(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })


  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |> 
      dplyr::collect() |> 
      sf::st_as_sf(
        coords = c("lon", "lat"), 
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_is_pentagon(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })


  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |> 
      dplyr::collect() |> 
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_is_pentagon(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })


  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct resolution for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to duckspatial_df
    test_data_sf <- test_data |> 
      dplyr::collect() |> 
      sf::st_as_sf(
        coords = c("lon", "lat"), 
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_is_pentagon("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })
  
})


testthat::test_that("pentagon returns true", {
  pentagon_df <- data.frame(h3string = "870800000ffffff")
  res <- ddbh3_is_pentagon(pentagon_df)
  ## Check false
  res_col <- dplyr::collect(res)
  expect_true(res_col$ispentagon)
})

testthat::test_that("hexagon returns false", {
  pentagon_df <- data.frame(h3string = "8ad02dcc1947fff")
  res <- ddbh3_is_pentagon(pentagon_df)
  ## Check false
  res_col <- dplyr::collect(res)
  expect_false(res_col$ispentagon)
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_is_pentagon() arguments work", {
  
  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |> 
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_is_pentagon(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ispentagon", colnames(res_col))
    expect_type(res_col$ispentagon, "logical")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_is_pentagon(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "logical")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_is_pentagon(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl"
    ))
    ## Apply operation with connection
    expect_no_message(ddbh3_is_pentagon(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_is_pentagon(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_is_pentagon(
      dplyr::collect(test_data), 
      new_column = "res",
      conn = conn_test, 
      name = "test_data_tbl",
      overwrite = TRUE
    ))

  })
  
})

## 1.3. Errors on weird inputs -----------


describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))
    
  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_is_pentagon(test_data, h3 = NULL))
    expect_error(ddbh3_is_pentagon(test_data, h3 = TRUE))
    expect_error(ddbh3_is_pentagon(test_data, h3 = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_is_pentagon(test_data, new_column = NULL))
    expect_error(ddbh3_is_pentagon(test_data, new_column = FALSE))
    expect_error(ddbh3_is_pentagon(test_data, new_column = 25))
  })
  
  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_is_pentagon("test_data_tbl", conn = NULL)))
    )
  })
  
  it("validates x argument type", {
    expect_error(ddbh3_is_pentagon(x = 999))
  })
  
  it("validates conn argument type", {
    expect_error(ddbh3_is_pentagon(test_data, conn = 999))
  })
  
  it("validates overwrite argument type", {
    expect_error(ddbh3_is_pentagon(test_data, overwrite = 999))
  })
  
  it("validates quiet argument type", {
    expect_error(ddbh3_is_pentagon(test_data, quiet = 999))
  })
  
  it("validates table name exists", {
    expect_error(ddbh3_is_pentagon(x = "999", conn = conn_test))
  })
  
  it("requires name to be single character string", {
    expect_error(ddbh3_is_pentagon(test_data, conn = conn_test, name = c('banana', 'banana')))
  })
})
  

# 2. is_h3() ---------------------------------------------------------

## 2.1. Input data in different formats ----------

testthat::describe("ddbh3_is_h3() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_is_h3(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_is_h3(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_is_h3(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_is_h3(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_is_h3("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

})


testthat::test_that("valid h3 cell returns true", {
  valid_df <- data.frame(h3string = "8ad02dcc1947fff")
  res <- ddbh3_is_h3(valid_df)
  res_col <- dplyr::collect(res)
  expect_true(res_col$ish3)
})

testthat::test_that("invalid h3 string returns false", {
  invalid_df <- data.frame(h3string = "not_an_h3_string")
  res <- ddbh3_is_h3(invalid_df)
  res_col <- dplyr::collect(res)
  expect_false(res_col$ish3)
})


## 2.2. Arguments work ------------

testthat::describe("ddbh3_is_h3() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_is_h3(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("ish3", colnames(res_col))
    expect_type(res_col$ish3, "logical")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_is_h3(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "logical")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_is_h3(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_is_h3(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_is_h3(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_is_h3(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 2.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_is_h3(test_data, h3 = NULL))
    expect_error(ddbh3_is_h3(test_data, h3 = TRUE))
    expect_error(ddbh3_is_h3(test_data, h3 = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_is_h3(test_data, new_column = NULL))
    expect_error(ddbh3_is_h3(test_data, new_column = FALSE))
    expect_error(ddbh3_is_h3(test_data, new_column = 25))
  })

  it("validates x argument type", {
    expect_error(ddbh3_is_h3(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_is_h3(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_is_h3(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_is_h3(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_is_h3(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_is_h3(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 3. is_res_class_iii() ---------------------------------------------------------

## 3.1. Input data in different formats ----------

testthat::describe("ddbh3_is_res_class_iii() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_is_res_class_iii(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_is_res_class_iii(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_is_res_class_iii(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_is_res_class_iii(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_is_res_class_iii("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

})


testthat::test_that("class III resolution returns true", {
  # Odd resolutions (1, 3, 5, 7, 9, 11, 13, 15) are Class III
  class_iii_df <- data.frame(h3string = "85283473fffffff")
  res <- ddbh3_is_res_class_iii(class_iii_df)
  res_col <- dplyr::collect(res)
  expect_true(res_col$isclassiii)
})

testthat::test_that("class II resolution returns false", {
  # Even resolutions (0, 2, 4, 6, 8, 10, 12, 14) are Class II
  class_ii_df <- data.frame(h3string = "8208000000fffff")
  res <- ddbh3_is_res_class_iii(class_ii_df)
  res_col <- dplyr::collect(res)
  expect_false(res_col$isclassiii)
})


## 3.2. Arguments work ------------

testthat::describe("ddbh3_is_res_class_iii() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_is_res_class_iii(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isclassiii", colnames(res_col))
    expect_type(res_col$isclassiii, "logical")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_is_res_class_iii(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "logical")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_is_res_class_iii(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_is_res_class_iii(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_is_res_class_iii(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_is_res_class_iii(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 3.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_is_res_class_iii(test_data, h3 = NULL))
    expect_error(ddbh3_is_res_class_iii(test_data, h3 = TRUE))
    expect_error(ddbh3_is_res_class_iii(test_data, h3 = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_is_res_class_iii(test_data, new_column = NULL))
    expect_error(ddbh3_is_res_class_iii(test_data, new_column = FALSE))
    expect_error(ddbh3_is_res_class_iii(test_data, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_is_res_class_iii("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_is_res_class_iii(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_is_res_class_iii(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_is_res_class_iii(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_is_res_class_iii(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_is_res_class_iii(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_is_res_class_iii(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})

# 4. is_vertex() ---------------------------------------------------------

## 4.1. Input data in different formats ----------

testthat::describe("ddbh3_is_vertex() works in different formats", {

  test_data_vertex <- ddbh3_h3_to_vertex(test_data)

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_is_vertex(test_data_vertex)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data_vertex)
    res <- ddbh3_is_vertex(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_is_vertex(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data_vertex |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_is_vertex(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_is_vertex("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

})


testthat::test_that("valid vertex returns true", {
  valid_vertex_df <- data.frame(h3vertex = "2222597fffffffff")
  res <- ddbh3_is_vertex(valid_vertex_df)
  res_col <- dplyr::collect(res)
  expect_true(res_col$isvertex)
})

testthat::test_that("invalid vertex returns false", {
  invalid_vertex_df <- data.frame(h3vertex = "not_a_vertex")
  res <- ddbh3_is_vertex(invalid_vertex_df)
  res_col <- dplyr::collect(res)
  expect_false(res_col$isvertex)
})


## 4.2. Arguments work ------------

testthat::describe("ddbh3_is_vertex() arguments work", {

  test_data_vertex <- ddbh3_h3_to_vertex(test_data)

  ## ARGUMENT 1 - H3VERTEX
  testthat::it("h3vertex argument works", {
    ## Rename h3vertex column
    test_data_mod <- test_data_vertex |>
      dplyr::rename(vertex = h3vertex)
    ## Apply operation with new h3vertex column name
    res <- ddbh3_is_vertex(test_data_mod, h3vertex = "vertex")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("isvertex", colnames(res_col))
    expect_type(res_col$isvertex, "logical")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_is_vertex(test_data_vertex, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "logical")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_is_vertex(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_is_vertex(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_is_vertex(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_is_vertex(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  test_data_vertex <- ddbh3_h3_to_vertex(test_data)

  ## Get h3vertex strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data_vertex))

  ## Tests
  it("requires h3vertex argument as character", {
    expect_error(ddbh3_is_vertex(test_data_vertex, h3vertex = NULL))
    expect_error(ddbh3_is_vertex(test_data_vertex, h3vertex = TRUE))
    expect_error(ddbh3_is_vertex(test_data_vertex, h3vertex = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_is_vertex(test_data_vertex, new_column = NULL))
    expect_error(ddbh3_is_vertex(test_data_vertex, new_column = FALSE))
    expect_error(ddbh3_is_vertex(test_data_vertex, new_column = 25))
  })

  it("validates x argument type", {
    expect_error(ddbh3_is_vertex(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_is_vertex(test_data_vertex, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_is_vertex(test_data_vertex, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_is_vertex(test_data_vertex, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_is_vertex(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_is_vertex(test_data_vertex, conn = conn_test, name = c('banana', 'banana')))
  })

})

#' @param nested Logical. If `TRUE`, children are returned as a nested list
#'   column (one row per parent cell). If `FALSE` (default), the result is
#'   unnested so each child cell occupies its own row.

#' @param output Character. Controls the return type. Options:
#'   \itemize{
#'     \item \code{"duckspatial_df"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
#'     \item \code{"sf"}: Eagerly collected sf object (uses memory)
#'     \item \code{"tibble"}: Eagerly collected tibble without geometry
#'     \item \code{"raw"}: Eagerly collected tibble with WKB geometry (list of raw vectors)
#'     \item \code{"geoarrow"}: Eagerly collected tibble with geoarrow geometry (geoarrow_vctr)
#'   }
#'   Can be set globally via \code{\link{ddbs_options}(output_type = "...")} or
#'   per-function via this argument. Per-function overrides global setting.

#' @param threads Integer. Number of threads to use. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use all available cores it detects (warning, on some shared HPC nodes the detected number of cores might be total number of cores on the node, not the per-job allocation).

#' @param name A character string of length one specifying the name of the table,
#'        or a character string of length two specifying the schema and table
#'        names. If `NULL` (the default), the function returns the result as an
#'        \code{sf} object

# skip tests on CRAN because they take too much time
# skip_if(Sys.getenv("TEST_ONE") != "")
# testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## Try to limit threads
Sys.setenv("OMP_THREAD_LIMIT" = 2)

# Setup default connection for tests
ddbh3_default_conn()

#' Convert longitude and latitude to H3 cell representations
#'
#' Convert geographic coordinates (longitude and latitude) into H3 cell
#' representations at a specified resolution, with different output formats
#'
#' @template x
#' @template lon
#' @template lat
#' @template resolution
#' @template new_column
#' @param h3_format Character. The format of the H3 cell index: `string` or
#' `bigint`
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#' 
#' @details
#' The three functions differ only in the output format of the H3 cell index:
#' * `ddbh3_lonlat_to_h3()` returns H3 cell indexes as strings (e.g. `"8928308280fffff"`) 
#' or as unsigned 64-bit integers (`UBIGINT`)
#' * `ddbh3_lonlat_to_spatial()` returns H3 cells as spatial hexagon polygons
#'
#' @template desc_formats
#' 
#' @template returns_tbl
#'
#' @name ddbh3_lonlat_to
#' @rdname ddbh3_lonlat_to
#' @aliases ddbh3_lonlat_to_h3 ddbh3_lonlat_to_spatial
#'
#' @examples
#' \donttest{
#' ## Load needed packages
#' library(duckdb)
#' library(duckh3)
#' library(duckspatial)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Create a connection with spatial and h3 extensions
#' conn <- ddbh3_create_conn(threads = 1)
#' 
#' ## TO H3 ------------
#' 
#' ## Add h3 strings as a new column (res 5)
#' points_strings_5_tbl <- ddbh3_lonlat_to_h3(
#'   points_tbl,
#'   resolution = 5
#' )
#' 
#' ## Add h3 UBIGINT as a new column (res 8 by default)
#' points_bigint_8_tbl <- ddbh3_lonlat_to_h3(
#'   points_tbl,
#'   new_column = "h3bigint",
#'   h3_format  = "bigint"
#' )
#' 
#' ## If column names are different from lon/lat:
#' points_renamed <- rename(points_tbl, long = lon, lati = lat)
#' 
#' ddbh3_lonlat_to_h3(
#'   points_renamed,
#'   lon = "long",
#'   lat = "lati",
#'   resolution = 10
#' )
#' 
#' 
#' ## Create a new table in the connection
#' ddbh3_lonlat_to_h3(
#'   points_tbl,
#'   conn = conn,
#'   name = "points_strings_8"
#' )
#' 
#' ## Open the created table lazily
#' points_lazy <- dplyr::tbl(conn, "points_strings_8")
#' 
#' ## Read it in memory
#' points_eager <- dbReadTable(conn, "points_strings_8")
#' 
#' 
#' ## TO SPATIAL -----------
#' 
#' ## Add h3 strings as a new column (res 5)
#' points_5_ddbs <- ddbh3_lonlat_to_spatial(
#'   points_tbl,
#'   resolution = 5
#' )
#' 
#' ## Create a new table in the connection
#' ddbh3_lonlat_to_spatial(
#'   points_tbl,
#'   conn = conn,
#'   name = "points_strings_spatial"
#' )
#' 
#' ## Open the created table lazily
#' as_duckspatial_df("points_strings_spatial", conn)
#' 
#' ## Read it in memory as an sf object
#' ddbs_read_table(conn, "points_strings_spatial")
#' }
NULL





#' @rdname ddbh3_lonlat_to
#' @export
ddbh3_lonlat_to_spatial <- function(
    x,
    lon = "lon",
    lat = "lat",
    resolution = 8,
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {


  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(lon, "lon")
  duckspatial:::assert_character_scalar(lat, "lat")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  built_fun <- glue::glue("
    ST_GeomFromWKB(
      h3_cell_to_boundary_wkb(
        h3_latlng_to_cell_string({lat}, {lon}, {resolution})
      )
    )"
  )

  # 2. Pass to template
  template_h3_to_spatial(
    x = x,
    conn = conn,
    name = name,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun
  ) 

}




#' @rdname ddbh3_lonlat_to
#' @export
ddbh3_lonlat_to_h3 <- function(
    x,
    lon = "lon",
    lat = "lat",
    resolution = 8,
    new_column = "h3string",
    h3_format = "string",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {


  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(lon, "lon")
  duckspatial:::assert_character_scalar(lat, "lat")
  duckspatial:::assert_integer_scalar(resolution, "resolution")
  duckspatial:::assert_numeric_interval(resolution, 0, 15, "resolution")

  # 1. Build parameters string
  built_fun <- switch(
    h3_format,
    "string" = glue::glue("h3_latlng_to_cell_string({lat}, {lon}, {resolution})"),
    "bigint" = glue::glue("h3_latlng_to_cell({lat}, {lon}, {resolution})"),
    cli::cli_abort("The {.arg h3_format} is not valid. Valid options: {.val {c('string', 'bigint')}}")
  )

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun
  ) 

}

#' Convert H3 cell indexes to vertex representations
#'
#' Convert H3 cell indexes stored as strings or unsigned 64-bit integers
#' (`UBIGINT`) to their vertex representations
#'
#'
#' @template x
#' @param n Integer. Vertex number to retrieve. Must be in the range 0–5 for
#'   hexagons and 0–4 for pentagons. Only used in `ddbh3_h3_to_vertex()`.
#' @template h3
#' @template h3vertex
#' @template new_column
#' @template conn_null
#' @template name
#' @template nested
#' @template overwrite
#' @template quiet
#' 
#' @details
#' The functions cover the full vertex workflow:
#' * `ddbh3_h3_to_vertex()` returns a single vertex of an H3 cell, identified
#'   by its vertex number `n` (0–5 for hexagons, 0–4 for pentagons), as an H3
#'   vertex string
#' * `ddbh3_h3_to_vertexes()` returns all vertices of an H3 cell as H3 vertex
#'   strings — either nested (one row per cell) or unnested (one row per vertex)
#'   depending on `nested`
#' * `ddbh3_vertex_to_lat()` returns the latitude of an H3 vertex string
#' * `ddbh3_vertex_to_lon()` returns the longitude of an H3 vertex string
#' * `ddbh3_vertex_to_spatial()` converts H3 vertex strings to spatial point
#'   geometries. If the input column is nested, vertices are automatically
#'   unnested and aggregated into a `MULTIPOINT` geometry per cell
#'
#' @template returns_tbl
#'
#' @name ddbh3_vertex
#' @aliases ddbh3_h3_to_vertex ddbh3_h3_to_vertexes ddbh3_vertex_to_spatial
#'
#' @examples
#' ## Load needed packages
#' library(duckh3)
#' library(duckspatial)
#' library(dplyr)
#' 
#' ## Setup the default connection with h3 and spatial extensions
#' ## This is a mandatory step to use duckh3 functions
#' ddbh3_default_conn(threads = 1)
#' 
#' ## Load example data
#' points_tbl <- read.csv(
#'   system.file("extdata/example_pts.csv", package = "duckh3")
#' )
#' 
#' ## Add h3 strings
#' points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)
#' 
#' ## TO VERTEX ---------------
#' 
#' ## Add second vertex
#' vertex_2_tbl <- ddbh3_h3_to_vertex(points_tbl, n = 2)
#' 
#' ## Add add vertexes (unnested)
#' vertexes_tbl <- ddbh3_h3_to_vertexes(points_tbl)
#' 
#' ## Add add vertexes (nested)
#' vertexes_nested_tbl <- ddbh3_h3_to_vertexes(points_tbl, nested = TRUE)
#' 
#' ## Add some vertexes with with mutate
#' points_tbl |> 
#'   mutate(
#'     v1 = ddbh3_h3_to_vertex(h3string, 1),
#'     v3 = ddbh3_h3_to_vertex(h3string, 3)
#'   )
#' 
#' ## VERTEX TO LON/LAT ------
#' 
#' ## Add coords
#' coords_vertex_tbl <- vertex_2_tbl |> 
#'   ddbh3_vertex_to_lon(new_column = "lon_v2") |> 
#'   ddbh3_vertex_to_lat(new_column = "lat_v2")
#' 
#' ## Add coords in mutate
#' vertex_2_tbl |> 
#'   mutate(
#'     lon_v2 = ddbh3_vertex_to_lon(h3vertex),
#'     lat_v2 = ddbh3_vertex_to_lat(h3vertex)
#'   )
#' 
#' ## VERTEX TO SPATIAL ------
#' 
#' ## Convert unnested vertexes (returns POINTS)
#' ddbh3_h3_to_vertexes(points_tbl) |> 
#'   ddbh3_vertex_to_spatial()
#' 
#' 
#' ## Convert nested vertexes (returns MULTIPOINTS)
#' ddbh3_h3_to_vertexes(points_tbl, nested = TRUE) |> 
#'   ddbh3_vertex_to_spatial()
NULL




#' @rdname ddbh3_vertex
#' @export
ddbh3_h3_to_vertex <- function(
    x,
    n = 0,
    h3 = "h3string",
    new_column = "h3vertex",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_integer_scalar(n, "n")
  duckspatial:::assert_numeric_interval(n, 0, 5, "n")


  # 1. Build parameters string
  built_fun <- glue::glue("h3_cell_to_vertex({h3}, {n})")


  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = glue::glue("h3_cell_to_vertex(x, {n})")
  ) 

}



#' @rdname ddbh3_vertex
#' @export
ddbh3_vertex_to_lon <- function(
    x,
    h3vertex = "h3vertex",
    new_column = "lon_vertex",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3vertex, "h3vertex")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_vertex_to_lng({h3vertex})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_vertex_to_lng(x)"
  )

}





#' @rdname ddbh3_vertex
#' @export
ddbh3_vertex_to_lat <- function(
    x,
    h3vertex = "h3vertex",
    new_column = "lat_vertex",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3vertex, "h3vertex")

  # 1. Build parameters string
  built_fun <- glue::glue("h3_vertex_to_lat({h3vertex})")

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_vertex_to_lat(x)"
  ) 

}





#' @rdname ddbh3_vertex
#' @export
ddbh3_h3_to_vertexes <- function(
    x,
    h3 = "h3string",
    new_column = "h3vertex",
    conn = NULL,
    name = NULL,
    nested = FALSE,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3, "h3")
  duckspatial:::assert_logic(nested, "nested")

  # 1. Build parameters string
  if (isTRUE(nested)) {
    built_fun <- glue::glue("h3_cell_to_vertexes({h3})")
  } else {
    built_fun <- glue::glue("UNNEST(h3_cell_to_vertexes({h3}))")
  }

  # 2. Pass to template
  template_h3_base(
    x = x,
    conn = conn,
    name = name,
    new_column = new_column,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun,
    base_fun = "h3_cell_to_vertexes(x)"
  ) 

}





#' @rdname ddbh3_vertex
#' @export
ddbh3_vertex_to_spatial <- function(
    x,
    h3vertex = "h3vertex",
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {
  
  
  # 0. Handle function-specific errors
  duckspatial:::assert_character_scalar(h3vertex, "h3vertex")


  # 1. Build parameters string

  ## 1.1. Check if it's nested
  is_nested <- check_nested_column(x, h3vertex)
  if (is_nested) {
  built_fun <- glue::glue("
    ST_POINT(
      h3_vertex_to_lng(UNNEST({h3vertex})),
      h3_vertex_to_lat(UNNEST({h3vertex}))
    )  
  ")
  } else {
  built_fun <- glue::glue("
    ST_POINT(
      h3_vertex_to_lng({h3vertex}),
      h3_vertex_to_lat({h3vertex})
    )  
  ")
  }


  # 2. Pass to template
  result <- template_h3_to_spatial(
    x = x,
    conn = conn,
    name = name,
    overwrite = overwrite,
    quiet = quiet,
    fun = built_fun
  )    


  # 3. Union in multipoint if it's nested
  if (is_nested) {
    result_crs <- result |> 
      dplyr::mutate(crs_duckspatial = "EPSG:4326")
    cols <- setdiff(colnames(result_crs), c("crs_duckspatial", "geometry"))
    duckspatial::ddbs_union_agg(result_crs, cols)
  } else {
    result
  }

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("duckspatial")

## create duckdb connection
conn_test <- duckh3::ddbh3_create_conn()

## Load example data
test_data <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
) |> 
  ddbh3_lonlat_to_h3(resolution = 10)

## Get data with vertex
test_data_vertex <- ddbh3_h3_to_vertex(test_data)
test_data_nested <- ddbh3_h3_to_vertexes(test_data, nested = TRUE)

# 1. h3_to_vertex() ---------------------------------------------------------

## 1.1. Input data in different formats ----------

testthat::describe("ddbh3_h3_to_vertex() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_h3_to_vertex(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_h3_to_vertex(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_h3_to_vertex(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_h3_to_vertex(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_h3_to_vertex("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

})


testthat::test_that("returns a valid vertex string", {
  cell_df <- data.frame(h3string = "8ad02dcc1947fff")
  res <- ddbh3_h3_to_vertex(cell_df, n = 0)
  res_col <- dplyr::collect(res)
  expect_type(res_col$h3vertex, "character")
  expect_true(nchar(res_col$h3vertex) > 0)
})

testthat::test_that("different vertex indices return different vertices", {
  cell_df <- data.frame(h3string = "8ad02dcc1947fff")
  res_0 <- dplyr::collect(ddbh3_h3_to_vertex(cell_df, n = 0))
  res_1 <- dplyr::collect(ddbh3_h3_to_vertex(cell_df, n = 1))
  expect_false(res_0$h3vertex == res_1$h3vertex)
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_h3_to_vertex() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_h3_to_vertex(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## ARGUMENT 2 - N
  testthat::it("n argument works", {
    ## All valid vertex indices return results
    for (i in 0:5) {
      res <- ddbh3_h3_to_vertex(test_data, n = i)
      res_col <- dplyr::collect(res)
      expect_type(res_col$h3vertex, "character")
    }
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_h3_to_vertex(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_h3_to_vertex(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_h3_to_vertex(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_h3_to_vertex(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_h3_to_vertex(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 1.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_h3_to_vertex(test_data, h3 = NULL))
    expect_error(ddbh3_h3_to_vertex(test_data, h3 = TRUE))
    expect_error(ddbh3_h3_to_vertex(test_data, h3 = 2))
  })

  it("requires n to be an integer scalar", {
    expect_error(ddbh3_h3_to_vertex(test_data, n = NULL))
    expect_error(ddbh3_h3_to_vertex(test_data, n = "a"))
    expect_error(ddbh3_h3_to_vertex(test_data, n = 1.5))
    expect_error(ddbh3_h3_to_vertex(test_data, n = c(1, 2)))
  })

  it("requires n to be in the range 0-5", {
    expect_error(ddbh3_h3_to_vertex(test_data, n = -1))
    expect_error(ddbh3_h3_to_vertex(test_data, n = 6))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_h3_to_vertex(test_data, new_column = NULL))
    expect_error(ddbh3_h3_to_vertex(test_data, new_column = FALSE))
    expect_error(ddbh3_h3_to_vertex(test_data, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_h3_to_vertex("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_h3_to_vertex(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_h3_to_vertex(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_h3_to_vertex(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_h3_to_vertex(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_h3_to_vertex(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_h3_to_vertex(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 2. vertex_to_lon() ---------------------------------------------------------

## 2.1. Input data in different formats ----------

testthat::describe("ddbh3_vertex_to_lon() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_vertex_to_lon(test_data_vertex)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data_vertex)
    res <- ddbh3_vertex_to_lon(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_vertex_to_lon(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data_vertex |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_vertex_to_lon(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_vertex_to_lon("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

})


testthat::test_that("returns a numeric longitude value", {
  vertex_df <- data.frame(h3vertex = "2222597fffffffff")
  res <- ddbh3_vertex_to_lon(vertex_df)
  res_col <- dplyr::collect(res)
  expect_type(res_col$lon_vertex, "double")
  expect_true(res_col$lon_vertex >= -180 & res_col$lon_vertex <= 180)
})


## 2.2. Arguments work ------------

testthat::describe("ddbh3_vertex_to_lon() arguments work", {

  ## ARGUMENT 1 - H3VERTEX
  testthat::it("h3vertex argument works", {
    ## Rename h3vertex column
    test_data_mod <- test_data_vertex |>
      dplyr::rename(vertex = h3vertex)
    ## Apply operation with new h3vertex column name
    res <- ddbh3_vertex_to_lon(test_data_mod, h3vertex = "vertex")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lon_vertex", colnames(res_col))
    expect_type(res_col$lon_vertex, "double")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_vertex_to_lon(test_data_vertex, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "double")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_vertex_to_lon(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_vertex_to_lon(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_vertex_to_lon(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_vertex_to_lon(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 2.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3vertex strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data_vertex))

  ## Tests
  it("requires h3vertex argument as character", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, h3vertex = NULL))
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, h3vertex = TRUE))
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, h3vertex = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, new_column = NULL))
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, new_column = FALSE))
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_vertex_to_lon("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_vertex_to_lon(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_vertex_to_lon(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_vertex_to_lon(test_data_vertex, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 3. vertex_to_lat() ---------------------------------------------------------

## 3.1. Input data in different formats ----------

testthat::describe("ddbh3_vertex_to_lat() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_vertex_to_lat(test_data_vertex)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data_vertex)
    res <- ddbh3_vertex_to_lat(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_vertex_to_lat(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data_vertex |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_vertex_to_lat(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_vertex_to_lat("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

})


testthat::test_that("returns a numeric latitude value", {
  vertex_df <- data.frame(h3vertex = "2222597fffffffff")
  res <- ddbh3_vertex_to_lat(vertex_df)
  res_col <- dplyr::collect(res)
  expect_type(res_col$lat_vertex, "double")
  expect_true(res_col$lat_vertex >= -90 & res_col$lat_vertex <= 90)
})


## 3.2. Arguments work ------------

testthat::describe("ddbh3_vertex_to_lat() arguments work", {

  ## ARGUMENT 1 - H3VERTEX
  testthat::it("h3vertex argument works", {
    ## Rename h3vertex column
    test_data_mod <- test_data_vertex |>
      dplyr::rename(vertex = h3vertex)
    ## Apply operation with new h3vertex column name
    res <- ddbh3_vertex_to_lat(test_data_mod, h3vertex = "vertex")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("lat_vertex", colnames(res_col))
    expect_type(res_col$lat_vertex, "double")
  })

  ## ARGUMENT 2 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_vertex_to_lat(test_data_vertex, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "double")
  })

  ## ARGUMENT 3 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_vertex_to_lat(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_vertex_to_lat(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_vertex_to_lat(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_vertex_to_lat(
      dplyr::collect(test_data_vertex),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 3.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3vertex strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data_vertex))

  ## Tests
  it("requires h3vertex argument as character", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, h3vertex = NULL))
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, h3vertex = TRUE))
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, h3vertex = 2))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, new_column = NULL))
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, new_column = FALSE))
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, new_column = 25))
  })

  it("requires connection when using table names", {
    expect_warning(
      expect_true(is.na(ddbh3_vertex_to_lat("test_data_tbl", conn = NULL)))
    )
  })

  it("validates x argument type", {
    expect_error(ddbh3_vertex_to_lat(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_vertex_to_lat(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_vertex_to_lat(test_data_vertex, conn = conn_test, name = c('banana', 'banana')))
  })

})


# 4. h3_to_vertexes() ---------------------------------------------------------

## 4.1. Input data in different formats ----------

testthat::describe("ddbh3_h3_to_vertexes() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {
    ## Check class
    res <- ddbh3_h3_to_vertexes(test_data)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {
    ## Check class
    test_data_df <- dplyr::collect(test_data)
    res <- ddbh3_h3_to_vertexes(test_data_df)
    expect_s3_class(res, "tbl_duckdb_connection")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Check class
    res <- ddbh3_h3_to_vertexes(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_h3_to_vertexes(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_h3_to_vertexes("sf_pts", conn = conn_test)
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

})


testthat::test_that("unnested result has more rows than nested", {
  ## Unnested (default): one row per vertex
  res_unnested <- ddbh3_h3_to_vertexes(test_data, nested = FALSE) |>
    dplyr::collect()
  ## Nested: one row per cell, vertex column contains a list
  res_nested <- ddbh3_h3_to_vertexes(test_data, nested = TRUE) |>
    dplyr::collect()
  expect_gt(nrow(res_unnested), nrow(res_nested))
})

testthat::test_that("nested result has same number of rows as input", {
  res_nested <- ddbh3_h3_to_vertexes(test_data, nested = TRUE) |>
    dplyr::collect()
  expect_equal(nrow(res_nested), nrow(dplyr::collect(test_data)))
})


## 1.2. Arguments work ------------

testthat::describe("ddbh3_h3_to_vertexes() arguments work", {

  ## ARGUMENT 1 - H3
  testthat::it("h3 argument works", {
    ## Rename h3string column
    test_data_mod <- test_data |>
      dplyr::rename(h3 = h3string)
    ## Apply operation with new h3 column name
    res <- ddbh3_h3_to_vertexes(test_data_mod, h3 = "h3")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("h3vertex", colnames(res_col))
    expect_type(res_col$h3vertex, "character")
  })

  ## ARGUMENT 2 - NESTED
  testthat::it("nested argument works", {
    ## nested = FALSE returns character column (unnested)
    res_false <- ddbh3_h3_to_vertexes(test_data, nested = FALSE)
    res_false_col <- dplyr::collect(res_false)
    expect_type(res_false_col$h3vertex, "character")
    ## nested = TRUE returns list column
    res_true <- ddbh3_h3_to_vertexes(test_data, nested = TRUE)
    res_true_col <- dplyr::collect(res_true)
    expect_type(res_true_col$h3vertex, "list")
  })

  ## ARGUMENT 3 - NEW_COLUMN
  testthat::it("new_column argument works", {
    res <- ddbh3_h3_to_vertexes(test_data, new_column = "res")
    expect_true("res" %in% colnames(res))
    res_col <- dplyr::collect(res)
    expect_in("res", colnames(res_col))
    expect_type(res_col$res, "character")
  })

  ## ARGUMENT 4 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_h3_to_vertexes(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_h3_to_vertexes(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_h3_to_vertexes(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_h3_to_vertexes(
      dplyr::collect(test_data),
      new_column = "res",
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 4.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data))

  ## Tests
  it("requires h3 argument as character", {
    expect_error(ddbh3_h3_to_vertexes(test_data, h3 = NULL))
    expect_error(ddbh3_h3_to_vertexes(test_data, h3 = TRUE))
    expect_error(ddbh3_h3_to_vertexes(test_data, h3 = 2))
  })

  it("requires nested argument as logical", {
    expect_error(ddbh3_h3_to_vertexes(test_data, nested = NULL))
    expect_error(ddbh3_h3_to_vertexes(test_data, nested = "yes"))
    expect_error(ddbh3_h3_to_vertexes(test_data, nested = 1))
  })

  it("requires new_column argument as character", {
    expect_error(ddbh3_h3_to_vertexes(test_data, new_column = NULL))
    expect_error(ddbh3_h3_to_vertexes(test_data, new_column = FALSE))
    expect_error(ddbh3_h3_to_vertexes(test_data, new_column = 25))
  })

  it("validates x argument type", {
    expect_error(ddbh3_h3_to_vertexes(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_h3_to_vertexes(test_data, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_h3_to_vertexes(test_data, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_h3_to_vertexes(test_data, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_h3_to_vertexes(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_h3_to_vertexes(test_data, conn = conn_test, name = c('banana', 'banana')))
  })

})




# 5. vertex_to_spatial() ---------------------------------------------------------

## 5.1. Input data in different formats ----------

testthat::describe("ddbh3_vertex_to_spatial() works in different formats", {

  ## FORMAT 1 - TBL_DUCKDB_CONNECTION
  testthat::it("returns the correct data for tbl_duckdb_connection", {

    ## Check nested
    res <- ddbh3_vertex_to_spatial(test_data_nested)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "MULTIPOINT"
    )

    ## Check unnested
    res <- ddbh3_vertex_to_spatial(test_data_vertex)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "POINT"
    )
  })

  ## FORMAT 2 - DATA.FRAME
  testthat::it("returns the correct data for data.frame", {

    ## Check nested
    test_data_df <- dplyr::collect(test_data_nested)
    res <- ddbh3_vertex_to_spatial(test_data_df)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "MULTIPOINT"
    )

    ## Check unnested
    test_data_df <- dplyr::collect(test_data_vertex)
    res <- ddbh3_vertex_to_spatial(test_data_df)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "POINT"
    )
  })

  ## FORMAT 3 - SF
  testthat::it("returns the correct data for sf", {
    
    ## Check nested
    ## Convert to sf
    test_data_sf <- test_data_nested |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Apply operation
    res <- ddbh3_vertex_to_spatial(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "MULTIPOINT"
    )

    ## Check unnested
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Apply operation
    res <- ddbh3_vertex_to_spatial(test_data_sf)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
    expect_all_equal(
      duckspatial::ddbs_geometry_type(res_col) |> as.character(),
      "POINT"
    )
  })

  ## FORMAT 4 - DUCKSPATIAL_DF
  testthat::it("returns the correct data for duckspatial_df", {
    ## Convert to duckspatial_df
    test_data_ddbs <- test_data_vertex |>
      dplyr::collect() |>
      duckspatial::ddbs_as_points(coords = c("lon", "lat"), crs = 4326)
    ## Check class
    res <- ddbh3_vertex_to_spatial(test_data_ddbs)
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

  ## FORMAT 5 - TABLE IN DUCKDB
  testthat::it("returns the correct data for table", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Convert to sf
    test_data_sf <- test_data_vertex |>
      dplyr::collect() |>
      sf::st_as_sf(
        coords = c("lon", "lat"),
        crs = 4326,
        remove = FALSE
      )
    ## Store table in connection
    duckspatial::ddbs_write_table(conn_test, test_data_sf, "sf_pts")
    ## Apply operation
    res <- ddbh3_vertex_to_spatial("sf_pts", conn = conn_test)
    ## Check class
    expect_s3_class(res, "duckspatial_df")
    ## Check type
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

})


testthat::test_that("unnested input returns one point per row", {
  ## Use flat vertex data (not nested)
  res <- ddbh3_vertex_to_spatial(test_data_vertex)
  res_col <- dplyr::collect(res)
  ## Same number of rows as input
  expect_equal(nrow(res_col), nrow(dplyr::collect(test_data_vertex)))
  ## Geometry column present
  expect_in("geometry", colnames(res_col))
})

testthat::test_that("nested input returns one multipoint per row", {
  ## Build nested vertex data from h3 cells
  test_data_nested <- ddbh3_h3_to_vertexes(test_data, nested = TRUE)
  res <- ddbh3_vertex_to_spatial(test_data_nested)
  res_col <- dplyr::collect(res)
  ## Same number of rows as original cell input
  expect_equal(nrow(res_col), nrow(dplyr::collect(test_data)))
  ## Geometry column present
  expect_in("geometry", colnames(res_col))
})


## 5.2. Arguments work ------------

testthat::describe("ddbh3_vertex_to_spatial() arguments work", {

  ## ARGUMENT 1 - H3VERTEX
  testthat::it("h3vertex argument works", {
    ## Rename h3vertex column
    test_data_mod <- test_data_vertex |>
      dplyr::rename(vertex = h3vertex)
    ## Apply operation with new h3vertex column name
    res <- ddbh3_vertex_to_spatial(test_data_mod, h3vertex = "vertex")
    ## Check class and geometry
    expect_s3_class(res, "duckspatial_df")
    res_col <- dplyr::collect(res)
    expect_in("geometry", colnames(res_col))
  })

  ## ARGUMENT 2 - DATABASE ARGUMENTS
  testthat::it("database arguments work", {
    ## Create connection
    conn_test <- ddbh3_create_conn()
    ## Apply operation with connection
    expect_message(ddbh3_vertex_to_spatial(
      dplyr::collect(test_data_vertex),
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Apply operation with connection, quiet
    expect_no_message(ddbh3_vertex_to_spatial(
      dplyr::collect(test_data_vertex),
      conn = conn_test,
      name = "test_data_tbl2",
      quiet = TRUE
    ))
    ## Doesn't overwrite existing table
    expect_error(ddbh3_vertex_to_spatial(
      dplyr::collect(test_data_vertex),
      conn = conn_test,
      name = "test_data_tbl"
    ))
    ## Overwrites existing table when overwrite = TRUE
    expect_true(ddbh3_vertex_to_spatial(
      dplyr::collect(test_data_vertex),
      conn = conn_test,
      name = "test_data_tbl",
      overwrite = TRUE
    ))
  })

})


## 5.3. Errors on weird inputs -----------

describe("errors", {

  ## Get h3vertex strings
  conn_test <- ddbh3_create_conn()
  duckdb::dbWriteTable(conn_test, "test_data_tbl", dplyr::collect(test_data_vertex))

  ## Tests
  it("requires h3vertex argument as character", {
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, h3vertex = NULL))
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, h3vertex = TRUE))
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, h3vertex = 2))
  })

  it("requires connection when using table names", {
    expect_error(ddbh3_vertex_to_spatial("test_data_tbl", conn = NULL))
  })

  it("validates x argument type", {
    expect_error(ddbh3_vertex_to_spatial(x = 999))
  })

  it("validates conn argument type", {
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, conn = 999))
  })

  it("validates overwrite argument type", {
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, overwrite = 999))
  })

  it("validates quiet argument type", {
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, quiet = 999))
  })

  it("validates table name exists", {
    expect_error(ddbh3_vertex_to_spatial(x = "999", conn = conn_test))
  })

  it("requires name to be single character string", {
    expect_error(ddbh3_vertex_to_spatial(test_data_vertex, conn = conn_test, name = c('banana', 'banana')))
  })

})

#' @param x Input data. One of:
#'   \describe{
#'     \item{`duckspatial_df`}{A lazy spatial data frame via dbplyr.}
#'     \item{`sf`}{A spatial data frame.}
#'     \item{`tbl_lazy`}{A lazy data frame from dbplyr.}
#'     \item{`data.frame`}{A standard R data frame.}
#'     \item{character string}{A table or view name in `conn`.}
#'     \item{character vector}{A vector of values to operate on in vectorized mode (requires `conn = NULL`).}
#'   }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_hierarchy.R
\name{ddbh3_get_hierarchy}
\alias{ddbh3_get_hierarchy}
\alias{ddbh3_get_parent}
\alias{ddbh3_get_center_child}
\alias{ddbh3_get_children}
\alias{ddbh3_get_n_children}
\title{Get parent and children H3 cells}
\usage{
ddbh3_get_parent(
  x,
  resolution = 8,
  h3 = "h3string",
  new_column = "h3parent",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_get_children(
  x,
  resolution = 8,
  h3 = "h3string",
  new_column = "h3children",
  conn = NULL,
  name = NULL,
  nested = FALSE,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_get_n_children(
  x,
  resolution = 8,
  h3 = "h3string",
  new_column = "h3n_children",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_get_center_child(
  x,
  resolution = 8,
  h3 = "h3string",
  new_column = "h3center_child",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{resolution}{A number specifying the resolution level of the H3
string (between 0 and 15)}

\item{h3}{The name of a column in \code{x} containing the H3 strings or
H3 unsigned 64-bit integers (\code{UBIGINT})}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{nested}{Logical. If \code{TRUE}, children are returned as a nested list
column (one row per parent cell). If \code{FALSE} (default), the result is
unnested so each child cell occupies its own row.}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Get the parent or children H3 cells of H3 cell indexes stored as strings or
unsigned 64-bit integers (\code{UBIGINT}) at a specified resolution:
\code{ddbh3_get_parent()}, \code{ddbh3_get_center_child()}, \code{ddbh3_get_children()},
and \code{ddbh3_get_n_children()}.
}
\details{
The four functions differ in the type of related cell they retrieve:
\itemize{
\item \code{ddbh3_get_parent()} returns the parent cell at a coarser resolution
\item \code{ddbh3_get_center_child()} returns the center child cell at a finer resolution
\item \code{ddbh3_get_children()} returns all children cells at a finer resolution
\item \code{ddbh3_get_n_children()} returns the number of children cells at a finer
resolution, without computing them
}
}
\examples{
## Load needed packages
library(duckh3)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Add h3 strings
points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)

## GET PARENTS ----------

## Get resolution-7 parent
points_parent_tbl <- ddbh3_get_parent(points_tbl, resolution = 7)

## Check the resolution
ddbh3_get_resolution(
  points_parent_tbl,
  h3 = "h3parent"
)

## Add with mutate
points_tbl |> 
  mutate(parent4 = ddbh3_get_parent(h3string, 4))

## GET CHILDREN ----------

## Get level 9 children
children_9_tbl <- ddbh3_get_children(points_tbl, resolution = 9)

## Get level 9 children in a nested list
children_9_nested_tbl <- ddbh3_get_children(points_tbl, resolution = 9, nested = TRUE)

## Add with mutate (nested)
points_tbl |> 
  mutate(children9 = ddbh3_get_children(h3string, 9))

## Add with mutate (unnested)
points_tbl |> 
  mutate(children9 = ddbh3_get_children(h3string, 9)) |> 
  mutate(children9 = unnest(children9))

## GET CENTER CHILD ------

## Get the center child of res 10 (1 child per row)
center_child_10_tbl <- ddbh3_get_center_child(points_tbl, resolution = 10)

## Add with mutate
points_tbl |> 
  mutate(center = ddbh3_get_center_child(h3string, 9))

## NUMBER OF CHILDREN -----

## How many children of level 10 does each level 8 have?
n_children_tbl <- ddbh3_get_n_children(points_tbl, resolution = 10)

## Add with mutate
points_tbl |> 
  mutate(n_children = ddbh3_get_n_children(h3string, 15))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_points_to.R
\name{ddbh3_points_to}
\alias{ddbh3_points_to}
\alias{ddbh3_points_to_h3}
\alias{ddbh3_points_to_spatial}
\title{Convert spatial points to H3 cell representations}
\usage{
ddbh3_points_to_spatial(
  x,
  resolution,
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_points_to_h3(
  x,
  resolution,
  new_column = "h3string",
  h3_format = "string",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{resolution}{A number specifying the resolution level of the H3
string (between 0 and 15)}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{h3_format}{Character. Output format for the H3 cell index. Either
\code{"string"} (default) or \code{"bigint"}. Only used in \code{ddbh3_points_to_h3()}.}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Convert spatial point geometries into H3 cell representations at a specified
resolution. Input must be an \code{duckspatial_df}, \code{sf}, or table with point
geometries
}
\details{
The two functions differ in the output format:
\itemize{
\item \code{ddbh3_points_to_h3()} returns the H3 cell index containing each point,
either as a string or \code{UBIGINT} depending on \code{h3_format}
\item \code{ddbh3_points_to_spatial()} returns the H3 cell hexagon polygon containing
each point as a spatial geometry
}
}
\examples{
## Load needed packages
library(duckh3)
library(duckspatial)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Convert to duckspatial_df
points_ddbs <- ddbs_as_points(points_tbl)

## TO H3 strings/ubigint ------------

## Add column with h3 strings at resolution 8
points_strings_ddbs <- ddbh3_points_to_h3(points_ddbs, 8)

## Add column with h3 ubigint at resolution 10
points_bigint_ddbs <- ddbh3_points_to_h3(
  points_ddbs, 
  resolution = 8,
  new_column = "h3bigint",
  h3_format = "bigint"
)

## TO SPATIAL -----------------

## Convert from POINTS to H3 POLYGONS of res 8
polygons_8_ddbs <- ddbh3_points_to_spatial(points_ddbs, 8)

## Collect as sf
polygons_8_sf <- ddbs_collect(polygons_8_ddbs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_h3_to.R
\name{ddbh3_h3_to}
\alias{ddbh3_h3_to}
\alias{ddbh3_h3_to_spatial}
\alias{ddbh3_h3_to_lon}
\alias{ddbh3_h3_to_lat}
\alias{ddbh3_strings_to_bigint}
\alias{ddbh3_bigint_to_strings}
\alias{ddbh3_h3_to_points}
\title{Convert H3 string or UBIGINT indexes to other representations}
\usage{
ddbh3_h3_to_lon(
  x,
  h3 = "h3string",
  new_column = "lon",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_h3_to_lat(
  x,
  h3 = "h3string",
  new_column = "lat",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_h3_to_spatial(
  x,
  h3 = "h3string",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_strings_to_bigint(
  x,
  h3 = "h3string",
  new_column = "h3bigint",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_bigint_to_strings(
  x,
  h3 = "h3bigint",
  new_column = "h3string",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_h3_to_points(
  x,
  h3 = "h3string",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{h3}{The name of a column in \code{x} containing the H3 strings or
H3 unsigned 64-bit integers (\code{UBIGINT})}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Convert H3 cell indexes stored as strings or UBIGINT into other
representations (e.g. \code{lon}, \code{lat}, \code{spatial})
}
\details{
The four functions differ only in the output format:
\itemize{
\item \code{ddbh3_h3_to_spatial()} converts H3 indexes to spatial hexagon polygons
\item \code{ddbh3_h3_to_lon()} extracts the longitude of the H3 cell centroid
\item \code{ddbh3_h3_to_lat()} extracts the latitude of the H3 cell centroid
\item \code{ddbh3_strings_to_bigint()} converts H3 indexes to unsigned 64-bit integers (\code{UBIGINT})
\item \code{ddbh3_bigint_to_strings()} converts H3 indexes to strings (e.g. \code{"8928308280fffff"})
\item \code{ddbh3_h3_to_points()} converts H3 indexes to spatial points located at the centroid
of the H3 cell
}
}
\examples{
## Load needed packages
library(duckh3)
library(duckspatial)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Add H3 string column
points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 6) |> 
  select(-lon, -lat)

## TO LON/LAT ------------

## Add longitude and latitude of the H3 string
points_coords_tbl <- points_tbl |> 
  ddbh3_h3_to_lat() |> 
  ddbh3_h3_to_lon()

## Add lon/lat with other names
points_coords_2_tbl <- points_tbl |> 
  ddbh3_h3_to_lat(new_column = "latitude") |> 
  ddbh3_h3_to_lon(new_column = "longitude")

## Add using mutate
points_tbl |> 
  mutate(
    lon = ddbh3_h3_to_lon(h3string),
    lat = ddbh3_h3_to_lat(h3string)
  )

## TO SPATIAL -----------------

## Convert h3 strings to spatial polygons
points_ddbs <- ddbh3_h3_to_spatial(points_tbl)

## Collect as sf
points_sf <- ddbs_collect(points_ddbs)

## FROM STRING TO UBIGINT -----

## Add ubigint, and remove strings
points_bigint_tbl <- ddbh3_strings_to_bigint(
  points_tbl, 
  new_column = "h3_integers"
) |> 
  select(-h3string)

## Add using mutate
points_tbl |> 
  mutate(h3int = ddbh3_strings_to_bigint(h3string))

## FROM UBIGINT TO STRING -----

## Add column with strings
points_strings_tbl <- ddbh3_bigint_to_strings(
  points_bigint_tbl, 
  h3 = "h3_integers"
) 

## Add using mutate
points_bigint_tbl |> 
  mutate(h3string = ddbh3_bigint_to_strings(h3_integers))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_extension.R
\name{ddbh3_default_conn}
\alias{ddbh3_default_conn}
\title{Get or create default DuckDB connection}
\usage{
ddbh3_default_conn(create = TRUE, upgrade_h3 = FALSE, ...)
}
\arguments{
\item{create}{Logical. If TRUE and no connection exists, create one.
Default is TRUE.}

\item{upgrade_h3}{Logical. If TRUE, will attempt to upgrade the h3 extension}

\item{...}{Additional parameters to pass to \code{\link[duckspatial:ddbs_create_conn]{duckspatial::ddbs_create_conn()}}}
}
\value{
Invisibly, the default \code{duckdb_connection}
}
\description{
Setup the default connection with h3 and spatial extensions installed and loaded.
It will be used internally by the package functions if no other connection is provided
}
\examples{
# Get or create default connection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_lonlat_to.R
\name{ddbh3_lonlat_to}
\alias{ddbh3_lonlat_to}
\alias{ddbh3_lonlat_to_h3}
\alias{ddbh3_lonlat_to_spatial}
\title{Convert longitude and latitude to H3 cell representations}
\usage{
ddbh3_lonlat_to_spatial(
  x,
  lon = "lon",
  lat = "lat",
  resolution = 8,
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_lonlat_to_h3(
  x,
  lon = "lon",
  lat = "lat",
  resolution = 8,
  new_column = "h3string",
  h3_format = "string",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{lon}{The name of a column in \code{x} containing the longitude}

\item{lat}{The name of a column in \code{x} containing the latitude}

\item{resolution}{A number specifying the resolution level of the H3
string (between 0 and 15)}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{h3_format}{Character. The format of the H3 cell index: \code{string} or
\code{bigint}}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Convert geographic coordinates (longitude and latitude) into H3 cell
representations at a specified resolution, with different output formats
}
\details{
The three functions differ only in the output format of the H3 cell index:
\itemize{
\item \code{ddbh3_lonlat_to_h3()} returns H3 cell indexes as strings (e.g. \code{"8928308280fffff"})
or as unsigned 64-bit integers (\code{UBIGINT})
\item \code{ddbh3_lonlat_to_spatial()} returns H3 cells as spatial hexagon polygons
}
}
\examples{
\donttest{
## Load needed packages
library(duckdb)
library(duckh3)
library(duckspatial)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Create a connection with spatial and h3 extensions
conn <- ddbh3_create_conn(threads = 1)

## TO H3 ------------

## Add h3 strings as a new column (res 5)
points_strings_5_tbl <- ddbh3_lonlat_to_h3(
  points_tbl,
  resolution = 5
)

## Add h3 UBIGINT as a new column (res 8 by default)
points_bigint_8_tbl <- ddbh3_lonlat_to_h3(
  points_tbl,
  new_column = "h3bigint",
  h3_format  = "bigint"
)

## If column names are different from lon/lat:
points_renamed <- rename(points_tbl, long = lon, lati = lat)

ddbh3_lonlat_to_h3(
  points_renamed,
  lon = "long",
  lat = "lati",
  resolution = 10
)


## Create a new table in the connection
ddbh3_lonlat_to_h3(
  points_tbl,
  conn = conn,
  name = "points_strings_8"
)

## Open the created table lazily
points_lazy <- dplyr::tbl(conn, "points_strings_8")

## Read it in memory
points_eager <- dbReadTable(conn, "points_strings_8")


## TO SPATIAL -----------

## Add h3 strings as a new column (res 5)
points_5_ddbs <- ddbh3_lonlat_to_spatial(
  points_tbl,
  resolution = 5
)

## Create a new table in the connection
ddbh3_lonlat_to_spatial(
  points_tbl,
  conn = conn,
  name = "points_strings_spatial"
)

## Open the created table lazily
as_duckspatial_df("points_strings_spatial", conn)

## Read it in memory as an sf object
ddbs_read_table(conn, "points_strings_spatial")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_hierarchy.R
\name{ddbh3_get_child_pos}
\alias{ddbh3_get_child_pos}
\title{Get the position of an H3 cell within its parent}
\usage{
ddbh3_get_child_pos(
  x,
  resolution = 8,
  h3 = "h3string",
  new_column = "h3child_pos",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{resolution}{A number specifying the resolution level of the H3
string (between 0 and 15)}

\item{h3}{The name of a column in \code{x} containing the H3 strings or
H3 unsigned 64-bit integers (\code{UBIGINT})}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Get the position of H3 cell indexes stored as strings or unsigned 64-bit
integers (\code{UBIGINT}) within their parent cell at a specified resolution.
The position is a zero-based index among all children of the parent cell.
}
\examples{
## Load needed packages
library(duckh3)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Add H3 string column
points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 6)

## Get position relative to resolution 4
ddbh3_get_child_pos(points_tbl, resolution = 4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_vertex.R
\name{ddbh3_vertex}
\alias{ddbh3_vertex}
\alias{ddbh3_h3_to_vertex}
\alias{ddbh3_h3_to_vertexes}
\alias{ddbh3_vertex_to_spatial}
\alias{ddbh3_vertex_to_lon}
\alias{ddbh3_vertex_to_lat}
\title{Convert H3 cell indexes to vertex representations}
\usage{
ddbh3_h3_to_vertex(
  x,
  n = 0,
  h3 = "h3string",
  new_column = "h3vertex",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_vertex_to_lon(
  x,
  h3vertex = "h3vertex",
  new_column = "lon_vertex",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_vertex_to_lat(
  x,
  h3vertex = "h3vertex",
  new_column = "lat_vertex",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_h3_to_vertexes(
  x,
  h3 = "h3string",
  new_column = "h3vertex",
  conn = NULL,
  name = NULL,
  nested = FALSE,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_vertex_to_spatial(
  x,
  h3vertex = "h3vertex",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{n}{Integer. Vertex number to retrieve. Must be in the range 0–5 for
hexagons and 0–4 for pentagons. Only used in \code{ddbh3_h3_to_vertex()}.}

\item{h3}{The name of a column in \code{x} containing the H3 strings or
H3 unsigned 64-bit integers (\code{UBIGINT})}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{h3vertex}{Name of the column containing H3 vertex strings. Defaults
to \code{"h3vertex"}}

\item{nested}{Logical. If \code{TRUE}, children are returned as a nested list
column (one row per parent cell). If \code{FALSE} (default), the result is
unnested so each child cell occupies its own row.}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Convert H3 cell indexes stored as strings or unsigned 64-bit integers
(\code{UBIGINT}) to their vertex representations
}
\details{
The functions cover the full vertex workflow:
\itemize{
\item \code{ddbh3_h3_to_vertex()} returns a single vertex of an H3 cell, identified
by its vertex number \code{n} (0–5 for hexagons, 0–4 for pentagons), as an H3
vertex string
\item \code{ddbh3_h3_to_vertexes()} returns all vertices of an H3 cell as H3 vertex
strings — either nested (one row per cell) or unnested (one row per vertex)
depending on \code{nested}
\item \code{ddbh3_vertex_to_lat()} returns the latitude of an H3 vertex string
\item \code{ddbh3_vertex_to_lon()} returns the longitude of an H3 vertex string
\item \code{ddbh3_vertex_to_spatial()} converts H3 vertex strings to spatial point
geometries. If the input column is nested, vertices are automatically
unnested and aggregated into a \code{MULTIPOINT} geometry per cell
}
}
\examples{
## Load needed packages
library(duckh3)
library(duckspatial)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Add h3 strings
points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)

## TO VERTEX ---------------

## Add second vertex
vertex_2_tbl <- ddbh3_h3_to_vertex(points_tbl, n = 2)

## Add add vertexes (unnested)
vertexes_tbl <- ddbh3_h3_to_vertexes(points_tbl)

## Add add vertexes (nested)
vertexes_nested_tbl <- ddbh3_h3_to_vertexes(points_tbl, nested = TRUE)

## Add some vertexes with with mutate
points_tbl |> 
  mutate(
    v1 = ddbh3_h3_to_vertex(h3string, 1),
    v3 = ddbh3_h3_to_vertex(h3string, 3)
  )

## VERTEX TO LON/LAT ------

## Add coords
coords_vertex_tbl <- vertex_2_tbl |> 
  ddbh3_vertex_to_lon(new_column = "lon_v2") |> 
  ddbh3_vertex_to_lat(new_column = "lat_v2")

## Add coords in mutate
vertex_2_tbl |> 
  mutate(
    lon_v2 = ddbh3_vertex_to_lon(h3vertex),
    lat_v2 = ddbh3_vertex_to_lat(h3vertex)
  )

## VERTEX TO SPATIAL ------

## Convert unnested vertexes (returns POINTS)
ddbh3_h3_to_vertexes(points_tbl) |> 
  ddbh3_vertex_to_spatial()


## Convert nested vertexes (returns MULTIPOINTS)
ddbh3_h3_to_vertexes(points_tbl, nested = TRUE) |> 
  ddbh3_vertex_to_spatial()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_extension.R
\name{ddbh3_create_conn}
\alias{ddbh3_create_conn}
\title{Create a DuckDB connection with spatial and h3 extensions}
\usage{
ddbh3_create_conn(
  dbdir = "memory",
  threads = NULL,
  memory_limit_gb = NULL,
  bigint = "integer64",
  ...
)
}
\arguments{
\item{dbdir}{String. Either "tempdir", "memory", or file path with .duckdb
or .db extension. Defaults to "memory".}

\item{threads}{Integer. Number of threads to use. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use all available cores it detects (warning, on some shared HPC nodes the detected number of cores might be total number of cores on the node, not the per-job allocation).}

\item{memory_limit_gb}{Numeric. Memory limit in GB. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use 80\% of available operating system memory it detects (warning, on some shared HPC nodes the detected memory might be the full node memory, not the per-job allocation).}

\item{bigint}{String. How to handle 64-bit integers. One of "integer64" or "numeric"}

\item{...}{Other parameters passed to \code{DBI::dbConnect()}}
}
\value{
A \code{duckdb_connection}
}
\description{
It creates a DuckDB connection, and then it installs and loads the
spatial and h3 extensions
}
\examples{
\donttest{
# load packages
library(duckspatial)
library(duckh3)

# create a duckdb database in memory
conn <- ddbh3_create_conn(dbdir = "memory", threads = 1)

# create an in-memory connection with 1 thread and 2GB memory limit
conn <- ddbh3_create_conn(threads = 1, memory_limit_gb = 2)

# Create a persistent database in disk
# conn <- ddbh3_create_conn(dbdir = "my_database.duckdb")

ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_is.R
\name{ddbh3_is}
\alias{ddbh3_is}
\alias{ddbh3_is_pentagon}
\alias{ddbh3_is_valid}
\alias{ddbh3_is_res_class_iii}
\alias{ddbh3_is_h3}
\alias{ddbh3_is_vertex}
\title{Check properties of H3 cell indexes}
\usage{
ddbh3_is_pentagon(
  x,
  h3 = "h3string",
  new_column = "ispentagon",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_is_h3(
  x,
  h3 = "h3string",
  new_column = "ish3",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_is_res_class_iii(
  x,
  h3 = "h3string",
  new_column = "isclassiii",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbh3_is_vertex(
  x,
  h3vertex = "h3vertex",
  new_column = "isvertex",
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input data. One of:
\describe{
\item{\code{duckspatial_df}}{A lazy spatial data frame via dbplyr.}
\item{\code{sf}}{A spatial data frame.}
\item{\code{tbl_lazy}}{A lazy data frame from dbplyr.}
\item{\code{data.frame}}{A standard R data frame.}
\item{character string}{A table or view name in \code{conn}.}
\item{character vector}{A vector of values to operate on in vectorized mode (requires \code{conn = NULL}).}
}}

\item{h3}{The name of a column in \code{x} containing the H3 strings or
H3 unsigned 64-bit integers (\code{UBIGINT})}

\item{new_column}{Name of the new column to create on the input data. If NULL,
the function will return a vector with the result}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{h3vertex}{Name of the column containing H3 vertex strings. Defaults
to \code{"h3vertex"}}
}
\value{
One of the following, depending on the inputs:
\describe{
\item{\code{tbl_lazy}}{If \code{x} is not spatial.}
\item{\code{duckspatial_df}}{If \code{x} is spatial (e.g. an \code{sf} or \code{duckspatial_df} object).}
\item{\code{TRUE} (invisibly)}{If \code{name} is provided, a table is created in the connection
and \code{TRUE} is returned invisibly.}
\item{vector}{If \code{x} is a character vector and \code{conn = NULL}, the function operates
in vectorized mode, returning a vector of the same length as \code{x}.}
}
}
\description{
Check properties of H3 cell indexes stored as strings or unsigned 64-bit
integers
}
\details{
The functions check different properties of H3 cell indexes, all
returning a logical column:
\itemize{
\item \code{ddbh3_is_pentagon()}: returns \code{TRUE} if the H3 cell is one of the 12
pentagonal cells that exist at each H3 resolution
\item \code{ddbh3_is_h3()}: returns \code{TRUE} if the H3 cell index is a valid H3 cell
\item \code{ddbh3_is_res_class_iii()}: returns \code{TRUE} if the H3 cell belongs to a
Class III resolution (odd resolutions: 1, 3, 5, 7, 9, 11, 13, 15)
\item \code{ddbh3_is_vertex()}: returns \code{TRUE} if the index is a valid H3 vertex
}
}
\examples{
## Load needed packages
library(duckh3)
library(dplyr)

## Setup the default connection with h3 and spatial extensions
## This is a mandatory step to use duckh3 functions
ddbh3_default_conn(threads = 1)

## Load example data
points_tbl <- read.csv(
  system.file("extdata/example_pts.csv", package = "duckh3")
)

## Add h3 strings
points_tbl <- ddbh3_lonlat_to_h3(points_tbl, resolution = 8)

## IS VALID H3 -------------

## Check if h3 indexes are valid
ddbh3_is_h3(points_tbl)

## Check in mutate
points_tbl |>
  mutate(valid = ddbh3_is_h3(h3string))

## IS PENTAGON -------------

## Check if h3 indexes are pentagons
ddbh3_is_pentagon(points_tbl)

## Check in mutate
points_tbl |>
  mutate(is_pent = ddbh3_is_pentagon(h3string))

## IS CLASS III ------------

## Check if h3 indexes belong to a Class III resolution
ddbh3_is_res_class_iii(points_tbl)

## Check across multiple resolutions
ddbh3_lonlat_to_h3(points_tbl, resolution = 7) |>
  ddbh3_is_res_class_iii()

## IS VERTEX ---------------

## Get vertexes first
vertex_tbl <- ddbh3_h3_to_vertex(points_tbl, n = 1)

## Check if indexes are valid vertexes
ddbh3_is_vertex(vertex_tbl, h3 = "h3vertex")

## Check in mutate (mix of h3 cells and vertexes)
vertex_tbl |>
  mutate(
    cell_valid  = ddbh3_is_h3(h3string),
    vertex_valid = ddbh3_is_vertex(h3vertex)
  )
}
